import unittest
import envi.cli as e_cli
import envi.memory as e_mem
import envi.memcanvas as e_mcanvas
import envi.memcanvas.renderers as e_render


class EnviCliTest(unittest.TestCase):
    def test_do_cmds(self):
        mem = e_mem.MemoryObject()
        mem.addMemoryMap(0x41410000, 7, 'testmem', b'@ABCDEF'*256)

        ecli = e_cli.EnviCli(mem)

        ecli.canvas = e_mcanvas.StringMemoryCanvas(mem)
        ecli.canvas.addRenderer("bytes", e_render.ByteRend())
        ecli.canvas.addRenderer("u_int_16", e_render.ShortRend())
        ecli.canvas.addRenderer("u_int_32", e_render.LongRend())
        ecli.canvas.addRenderer("u_int_64", e_render.QuadRend())

        # do_alias
        ecli.do_alias('')
        output = ecli.canvas.strval
        self.assertIn('Runtime Aliases (not saved):', output)
        self.assertIn('Configured Aliases:', output)
        ecli.canvas.clearCanvas()

        # do_binstr
        ecli.do_binstr('47145')
        output = ecli.canvas.strval
        self.assertIn('0x0000b829 (47145) 1011100000101001', output)
        ecli.canvas.clearCanvas()

        # do_clear
        ecli.vprint("foobar")
        ecli.do_clear('')
        output = ecli.canvas.strval
        self.assertEqual('', output)
        ecli.canvas.clearCanvas()

        # do_config
        ecli.do_config('')
        output = ecli.canvas.strval
        self.assertIn('cli.verbose = False\n', output)
        ecli.canvas.clearCanvas()

        # do_EOF
        ecli.do_EOF('')
        output = ecli.canvas.strval
        self.assertIn('Use quit\n', output)
        ecli.canvas.clearCanvas()

        # do_eval
        ecli.do_eval('0x40 + 0x47145')
        output = ecli.canvas.strval
        self.assertIn('0x40 + 0x47145 = 0x00047185 (291205)\n', output)
        ecli.canvas.clearCanvas()

        # do_maps
        ecli.do_maps('')
        output = ecli.canvas.strval
        self.assertIn('0x41410000', output)
        ecli.canvas.clearCanvas()

        ecli.do_maps('0x41410001')
        output = ecli.canvas.strval
        self.assertIn(' 0x41410001\n0x41410000', output)
        ecli.canvas.clearCanvas()

        # do_mem
        ecli.do_mem('0x41410001 20')
        output = ecli.canvas.strval
        self.assertIn('0x41410001  41 42 43 44 45 46 40 41 42 43 44 45 46 40 41 42   ABCDEF@ABCDEF@AB\n0x41410011  43 44 45 46 40 41 42 43 44 45 46 40 41 42 43 44   CDEF@ABCDEF@ABCD\n', output)
        ecli.canvas.clearCanvas()

        ecli.do_mem('-F u_int_16 0x41410001 20')
        output = ecli.canvas.strval
        self.assertIn('0x41410001  4241 4443 4645 4140 4342 4544 4046 4241   ABCDEF@ABCDEF@AB\n0x41410011  4443 4645 4140 4342 4544 4046 4241 4443   CDEF@ABCDEF@ABCD\n', output)
        ecli.canvas.clearCanvas()

        ecli.do_mem('-F u_int_32 0x41410001 20')
        output = ecli.canvas.strval
        self.assertIn('0x41410001  44434241 41404645 45444342 42414046   ABCDEF@ABCDEF@AB\n0x41410011  46454443 43424140 40464544 44434241   CDEF@ABCDEF@ABCD\n', output)
        ecli.canvas.clearCanvas()

        ecli.do_mem('-F u_int_64 0x41410001 20')
        output = ecli.canvas.strval
        self.assertIn('0x0000000041410001  4140464544434241 4241404645444342   ABCDEF@ABCDEF@AB\n0x0000000041410011  4342414046454443 4443424140464544   CDEF@ABCDEF@ABCD\n', output)
        ecli.canvas.clearCanvas()

        # do_memcmp
        ecli.do_memcmp('0x41410001 0x41410005 20')
        output = ecli.canvas.strval
        self.assertIn('==== 20 byte difference at offset 0\n0x41410001:4142434445464041424344454640414243444546\n0x41410005:4546404142434445464041424344454640414243\n', output)
        ecli.canvas.clearCanvas()
        
        # do_memdump
        ecli.do_memdump('0x41410001 /tmp/foo40 20')
        output = ecli.canvas.strval
        self.assertIn('wrote 20 bytes.\n', output)
        ecli.canvas.clearCanvas()
        with open('/tmp/foo40', 'rb') as testfile:
            self.assertEqual(testfile.read(), b'ABCDEF@ABCDEF@ABCDEF')
        ecli.canvas.clearCanvas()
       
        # do_python
        ecli.do_python('with open("/tmp/foo41", "wb") as testfile:\n\ttestfile.write(b"Test Worked")')
        output = ecli.canvas.strval
        with open('/tmp/foo41', 'rb') as testfile:
            self.assertEqual(testfile.read(), b'Test Worked')
        ecli.canvas.clearCanvas()

        # do_saveout
        ecli.do_saveout('/tmp/foo42 search ABCDEF')
        output = ecli.canvas.strval
        self.assertIn('\ndone (256 results).\n', output)
        ecli.canvas.clearCanvas()

        # do_search
        ecli.do_search('ABCDEF')
        output = ecli.canvas.strval
        self.assertIn('\ndone (256 results).\n', output)
        ecli.canvas.clearCanvas()




from vivisect.symboliks.common import *

from vivisect.const import *

class SymbolikEffect:
    '''
    A single symbolik effect...
    '''
    efftype = None

    def __init__(self, va):
        self.va = va

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        raise Exception('%s must implement __eq__!' % (self.__class__.__name__,))

    def reduce(self, emu=None):
        raise Exception('%s must implement reduce()!' % (self.__class__.__name__))

    # Needed for compatibility w/ SymbolikExpressionParser calling Symbolik{Effect, Base}
    def update(self, emu):
        return self.applyEffect(emu)

    def walkTree(self, cb, ctx=None, once=True):
        raise Exception('%s must implement walkTree()!' % (self.__class__.__name__))

    def applyEffect(self, emu):
        raise Exception('%s must implement applyEffect!' % (self.__class__.__name__,))

    def render(self, canvas, vw):
        canvas.addText(str(self))

class DebugEffect(SymbolikEffect):
    '''
    DebugEffect is used to represent an NOP effect that we want logged.
    Example: DebugEffect is created for instructions that are unsupported.

        return DebugEffect(op.va, "%s Needs %s" % (self.__class__.__name__, repr(op)))
    '''
    efftype = EFFTYPE_DEBUG

    def __init__(self, va, msg):
        SymbolikEffect.__init__(self, va)
        self.msg = msg

    def __repr__(self):
        return 'DebugEffect( 0x%.8x, %s )' % (self.va, self.msg)

    def __str__(self):
        return '%s' % self.msg

    def __eq__(self, other):
        if other is None:
            return False
        if self.__class__ != other.__class__:
            return False
        if not self.msg == other.msg:
            return False

        return True

    def walkTree(self, cb, ctx=None, once=True):
        pass

    def reduce(self, emu=None):
        pass

    def applyEffect(self, emu):
        return self

class SetVariable(SymbolikEffect):

    efftype = EFFTYPE_SETVAR

    def __init__(self, va, varname, symobj):
        SymbolikEffect.__init__(self, va)
        self.varname = varname
        self.symobj = symobj

    def __repr__(self):
        return 'SetVariable( 0x%.8x, %s, %s )' % (self.va, repr(self.varname), repr(self.symobj))

    def __str__(self):
        return '%s = %s' % (self.varname, str(self.symobj))

    def render(self, canvas, vw):
        canvas.addNameText(self.varname)
        canvas.addText(' = ')
        self.symobj.render(canvas, vw)

    def __eq__(self, other):
        if other is None:
            return False
        if self.__class__ != other.__class__:
            return False
        if not self.varname == other.varname:
            return False
        if not self.symobj == other.symobj:
            return False

        return True

    def walkTree(self, cb, ctx=None, once=True):
        self.symobj = self.symobj.walkTree(cb, ctx=ctx, once=once)

    def reduce(self, emu=None):
        self.symobj = self.symobj.reduce(emu=emu)

    def applyEffect(self, emu):
        symobj = self.symobj.update(emu)
        emu.setSymVariable(self.varname, symobj)
        return SetVariable(self.va, self.varname, symobj)

    def render(self, canvas, vw):
        rname = vw.arch._arch_dis._dis_regctx.getRealRegisterName(self.varname)
        canvas.addNameText(str(self.varname), name=rname, typename="registers")
        canvas.addText(" = %s" % str(self.symobj))

class ReadMemory(SymbolikEffect):

    efftype = EFFTYPE_READMEM

    def __init__(self, va, symaddr, symsize):
        SymbolikEffect.__init__(self, va)
        self.symaddr = symaddr
        self.symsize = symsize

    def __repr__(self):
        t = (self.va, repr(self.symaddr), repr(self.symsize))
        return 'ReadMemory( 0x%.8x, %s, %s )' % t

    def __str__(self):
        return '[ %s : %s ]' % (str(self.symaddr), str(self.symsize))

    def __eq__(self, other):
        if other is None:
            return False
        if self.__class__ != other.__class__:
            return False
        if not self.symaddr == other.symaddr:
            return False
        if not self.symsize == other.symsize:
            return False

        return True

    def walkTree(self, cb, ctx=None, once=True):
        self.symaddr = self.symaddr.walkTree(cb, ctx=ctx, once=once)
        self.symsize = self.symsize.walkTree(cb, ctx=ctx, once=once)

    def reduce(self, emu=None):
        self.symaddr = self.symaddr.reduce(emu=emu)
        self.symsize = self.symsize.reduce(emu=emu)

    def applyEffect(self, emu):
        symaddr = self.symaddr.update(emu)
        symsize = self.symsize.update(emu)
        return ReadMemory(self.va, symaddr, symsize)

class WriteMemory(SymbolikEffect):

    efftype = EFFTYPE_WRITEMEM

    def __init__(self, va, symaddr, symsize, symval):
        SymbolikEffect.__init__(self, va)
        self.symval = symval
        self.symaddr = symaddr
        self.symsize = symsize

    def __repr__(self):
        t = (self.va, repr(self.symaddr), repr(self.symsize), repr(self.symval))
        return 'WriteMemory( 0x%.8x, %s, %s, %s )' % t

    def __str__(self):
        t = (str(self.symaddr), str(self.symsize), str(self.symval))
        return '[ %s : %s ] = %s' % t

    def __eq__(self, other):
        if other is None:
            return False
        if self.__class__ != other.__class__:
            return False
        if not self.symval == other.symval:
            return False
        if not self.symaddr == other.symaddr:
            return False
        if not self.symsize == other.symsize:
            return False

        return True

    def walkTree(self, cb, ctx=None, once=True):
        self.symval = self.symval.walkTree(cb, ctx=ctx, once=once)
        self.symaddr = self.symaddr.walkTree(cb, ctx=ctx, once=once)
        self.symsize = self.symsize.walkTree(cb, ctx=ctx, once=once)

    def reduce(self, emu=None):
        self.symaddr = self.symaddr.reduce(emu=emu)
        self.symsize = self.symsize.reduce(emu=emu)
        self.symval = self.symval.reduce(emu=emu)

    def applyEffect(self, emu):
        symaddr = self.symaddr.update(emu)
        symsize = self.symsize.update(emu)
        symval = self.symval.update(emu)
        emu.writeSymMemory(symaddr, symval)
        return WriteMemory(self.va, symaddr, symsize, symval)

class CallFunction(SymbolikEffect):
    '''
    This effect represents a procedural branch.  They are recorded specially
    because they may have effect on the overall system even though their
    outputs are not stored.

    NOTE: argsyms will be None while we haven't had a definition to use..
    '''

    efftype = EFFTYPE_CALLFUNC

    def __init__(self, va, funcsym, argsyms=None):
        SymbolikEffect.__init__(self, va)
        self.funcsym = funcsym
        self.argsyms = argsyms

    def __repr__(self):
        return 'CallFunction( 0x%.8x, %s, %s )' % (self.va, repr(self.funcsym), repr(self.argsyms))

    def __str__(self):
        argstr = '?'
        if self.argsyms is not None:
            argstr = ','.join(str(x) for x in self.argsyms)
        return '%s(%s)' % (self.funcsym, argstr)

    def __eq__(self, other):
        if other is None:
            return False
        if self.__class__ != other.__class__:
            return False
        if not self.funcsym == other.funcsym:
            return False
        if not self.argsyms == other.argsyms:
            return False

        return True

    def walkTree(self, cb, ctx=None, once=True):
        self.funcsym = self.funcsym.walkTree(cb, ctx=ctx, once=once)
        if self.argsyms is not None:
            self.argsyms = [ a.walkTree(cb,ctx=ctx, once=once) for a in self.argsyms ]

    def reduce(self, emu=None):
        self.funcsym = self.funcsym.reduce(emu=emu)
        if self.argsyms is not None:
            self.argsyms = [ x.reduce(emu=emu) for x in self.argsyms ]

    def applyEffect(self, emu):
        emu.setMeta('calling_va', self.va)
        funcsym = self.funcsym.update(emu)

        # If we have argsyms, the function's work has been broken out
        # already (probably by applying effects once already...)
        if self.argsyms is not None:
            argsyms = [x.update(emu) for x in self.argsyms]
            return CallFunction(self.va, funcsym, argsyms)

        # Without argsyms, we are probably a call who is being applied to
        # an emulator for the first time!  Let the emulator handle it...
        argsyms = emu.applyFunctionCall(funcsym)
        return CallFunction(self.va, funcsym, argsyms)

    def render(self, canvas, vw):

        # The symbol for the function name...
        self.funcsym.render(canvas, vw)
        canvas.addText('(')

        if self.argsyms is None:
            canvas.addText('?')
        else:
            argmax = len(self.argsyms) - 1

            for i, argsym in enumerate(self.argsyms):
                argsym.render(canvas, vw)

                if i < argmax:
                    canvas.addText(',')

        canvas.addText(')')

class ConstrainPath(SymbolikEffect):

    efftype = EFFTYPE_CONSTRAIN

    def __init__(self, va, addrsym, cons):
        SymbolikEffect.__init__(self, va)
        self.addrsym = addrsym
        self.cons = cons

    def walkTree(self, cb, ctx=None, once=True):
        # TODO: Should be also walk the addrsym here?
        self.cons.walkTree(cb, ctx=ctx, once=once)

    def reduce(self, emu=None):
        self.addrsym = self.addrsym.reduce(emu=emu)
        self.cons = self.cons.reduce(emu=emu)

    def __repr__(self):
        return 'ConstrainPath( 0x%.8x, %s, %s )' % (self.va, repr(self.addrsym), repr(self.cons))

    def __str__(self):
        return 'if (%s)' % (str(self.cons),)

    def render(self, canvas, vw):
        canvas.addText('if (')
        self.cons.render(canvas, vw)
        canvas.addText(')')

    def __eq__(self, other):
        if other is None:
            return False
        if self.__class__ != other.__class__:
            return False
        if not self.addrsym == other.addrsym:
            return False
        if not self.cons == other.cons:
            return False

        return True

    def applyEffect(self, emu):
        addrsym = self.addrsym.update(emu)
        cons = self.cons.update(emu)
        return ConstrainPath(self.va, addrsym, cons)



import logging
import unittest

import vivisect
import envi.exc as e_exc

import vivisect.tests.utils as v_t_utils

from envi.const import *


logger = logging.getLogger(__name__)


class VivMemoryTest(v_t_utils.VivTest):
    @classmethod
    def setUpClass(cls):
        cls.vw = vivisect.VivWorkspace()
        cls.vw.addMemoryMap(0x1000, MM_EXEC|MM_READ|MM_WRITE, 'foo1', b'A'*0x1000)
        cls.vw.addMemoryMap(0x2000, MM_EXEC|MM_READ, 'foo2', b'B'*0x1000)
        cls.vw.addMemoryMap(0x3000, MM_READ, 'foo3', b'C'*0x1000)
        cls.vw.addMemoryMap(0x4100, MM_READ, 'foo4', b'C'*0x1000)

    def test_memory_fails(self):
        with self.assertRaises(e_exc.SegmentationViolation):
            self.vw._reqProbeMem(0x1000, 0x3000, MM_EXEC|MM_READ|MM_WRITE)
        with self.assertRaises(e_exc.SegmentationViolation):
            self.vw._reqProbeMem(0x3000, 0x1000, MM_EXEC|MM_READ)
        with self.assertRaises(e_exc.SegmentationViolation):
            self.vw._reqProbeMem(0x1000, 0x4000, MM_READ)
        with self.assertRaises(e_exc.SegmentationViolation):
            self.vw._reqProbeMem(0x3000, 0x1001, MM_READ)
        with self.assertRaises(e_exc.SegmentationViolation):
            self.vw._reqProbeMem(0x3000, 0x1200, MM_READ)
        with self.assertRaises(e_exc.SegmentationViolation):
            self.vw._reqProbeMem(0x400, 0x1000, MM_READ)
        with self.assertRaises(e_exc.SegmentationViolation):
            self.vw._reqProbeMem(0x100, 0x3ff0, MM_READ)
        with self.assertRaises(e_exc.SegmentationViolation):
            self.vw._reqProbeMem(0x3000, 0xff0, MM_READ|MM_WRITE)

    def test_memory_success(self):
        self.vw._reqProbeMem(0x1f00, 0x300,  MM_EXEC|MM_READ)
        self.vw._reqProbeMem(0x1000, 0x2000, MM_EXEC|MM_READ)
        self.vw._reqProbeMem(0x1000, 0x3000, MM_READ)
        self.vw._reqProbeMem(0x3000, 0x1000, MM_READ)




import random

tips = [
    'Hold <shift> to move using the mouse in Function Graph view!',
    'You may drag rows from List Views to Memory Views',
    'You can enter any valid python expression into the expression bar at the top'
]

def getRandomTip():
    return random.choice(tips)



import math
import envi.exc as e_exc

# tuples of (Exponent bitlen, Exponent Bias)
ieee754_formats = {
    16:  (5, 0xF),
    32:  (8, 0x7F),
    64:  (11, 0x3FF),
    80:  (15, 0x3FFF),  # non-implicit j bit
    # At least on intel, don't go beyond 80 bits (even that's suspect)
    # because floating point errors
    # just compound so mother trucking hard that it errors out into oblivion
    # 128: (15, 0x3FFF),
}


def float_decode(valu, bsize):
    '''
    https://software.intel.com/sites/default/files/managed/ae/01/floating-point-reference-sheet.pdf
    '''
    exp_len, exp_bias = ieee754_formats[bsize]

    loc = bsize - 1
    sign = (valu & (1 << loc)) >> loc

    loc -= exp_len
    mask = (2**exp_len) - 1
    exponent = (valu & (mask << loc)) >> loc

    if bsize == 80:  # BECAUSE INTEL HAS TO BE OH SO SPECIAL
        jbit_len = 1
        integer = (valu & (1 << (loc-1))) >> (loc-1)
    else:
        jbit_len = 0
        integer = 1 if exponent != 0 else 0

    frac_len = bsize - jbit_len - exp_len - 1
    mask = (2**frac_len) - 1
    fraction = valu & mask
    m = integer + (fraction / (2.0**frac_len))

    if exponent == 0:
        # denormal
        if fraction != 0:
            return ((-1)**sign) * m * (2**(1-exp_bias))
        else:
            return 0
    elif 1 <= exponent < ((2**exp_len) - 1):
        # normal
        return ((-1)**sign) * m * (2**(exponent-exp_bias))
    else:
        # various forms of infinity or NaN
        # The kind of NaN does matter
        highbit = (fraction & (1 << (frac_len-1))) >> (frac_len-1)
        if fraction == 0:
            raise e_exc.InvalidOperand(valu)
        elif highbit == 1:
            raise e_exc.SignalNaN()
        else:
            raise e_exc.QuietNaN()


def float_encode(valu, bsize):
    exp_len, exp_bias = ieee754_formats[bsize]

    # sign bit
    encoded = 1 if valu < 0 else 0
    if valu < 0:
        valu *= -1
        encoded = 1
    else:
        encoded = 0

    # find the exponent
    exponent = 0
    if valu >= 2:
        while valu >= 2:
            valu /= 2
            exponent += 1
    elif valu < 1:
        while valu < 1:
            valu *= 2
            exponent -= 1

    mask = (2**exp_len) - 1
    exponent += exp_bias
    exponent &= mask
    encoded <<= exp_len

    encoded |= exponent

    if bsize == 80:
        jbit_len = 1
        jbit = 1 if exponent != 0 else 0
        encoded <<= 1
        encoded |= jbit
    else:
        # the jbit is hidden in normal reprs of iee754
        jbit_len = 0

    valu %= 1

    frac_len = bsize - jbit_len - exp_len - 1
    for i in range(frac_len):
        encoded <<= 1
        valu *= 2
        if valu > 1:
            valu %= 1
            encoded |= 1
    return encoded



"""
PE Extended analysis module.
"""

import vivisect
import envi.bits as e_bits

def analyze(vw):
    """
    """
    # Go through the relocations and create locations for them
    for segva, segsize, segname, segfname in vw.getSegments():
        reloc_va = vw.getFileMeta(segfname, "reloc_va")
        # Found binaries with multiple sections named .reloc where one was iat another
        # was actual reloc
        if reloc_va != segva:
            continue

        offset, bytes = vw.getByteDef(segva)
        while offset < segsize:
            # error cehck to make sure we are providing four bytes
            # to the parse routine
            if len(bytes[offset+4:offset+8]) != 4:
                break

            basepage = e_bits.parsebytes(bytes, offset, 4)

            vaoff = segva + offset
            vw.makeNumber(vaoff, 4)
            vw.makeName(vaoff, "reloc_chunk_%.8x" % vaoff)

            recsize = e_bits.parsebytes(bytes, offset+4, 4)
            if offset + recsize > segsize:
                break
            vw.makeNumber(segva+offset+4, 4)

            ioff = offset + 8
            while ioff < offset+recsize:
                vw.makeNumber(segva + ioff, 2)
                ioff += 2

            offset += recsize
            if recsize == 0:
                break





import envi.exc as e_exc
import envi.archs.i386 as e_i386
import envi.archs.amd64 as e_amd64

import vivisect.impemu.emulator as v_i_emulator
import vivisect.impemu.platarch.i386 as v_i_i386
import vivisect.impemu.platarch.amd64 as v_i_amd64

from vivisect.impemu.emulator import imphook

class Linuxi386Emulator(v_i_i386.i386WorkspaceEmulator):
    def __init__(self, vw, **kwargs):
        v_i_i386.i386WorkspaceEmulator.__init__(self, vw, **kwargs)

    def i_int(self, op):
        syscall = op.getOperValue(0)
        if syscall != 0x80:
            raise e_exc.BreakpointHit(self)

        reg = self.getRegister(e_i386.REG_EAX)
        if reg == 1:
            self.stopEmu()

class LinuxAmd64Emulator(v_i_amd64.Amd64WorkspaceEmulator):
    def __init__(self, vw, **kwargs):
        v_i_amd64.Amd64WorkspaceEmulator.__init__(self, vw, **kwargs)

    def i_syscall(self, op):
        # TODO: there is a long list of >300 functions we'd need to deal with here
        # some of which would exit
        pass



import math
import random
import logging
import itertools
import traceback

import visgraph.layouts as vg_layouts

logger = logging.getLogger(__name__)

# coulomb's constant approximation
ke = 8.98755

def irand():
    return random.randint(100, 200)

class vector:

    def __init__(self, force=0.0, angle=0.0):
        self.force = force
        self.angle = angle

    def addDrag(self, drag):
        self.force *= drag

    def addVect(self, force, angle):
        '''
        Update the vector by adding the given force/angle.
        '''
        # Use head-to-tail addition
        dx = math.cos(self.angle) * self.force + math.cos(angle) * force
        dy = math.sin(self.angle) * self.force + math.sin(angle) * force
        self.force = math.hypot(dx,dy)
        self.angle = math.atan2(dy, dx)

    def getMovement(self, x, y, maxmov=None):
        dx = math.cos(self.angle) * self.force
        dy = math.sin(self.angle) * self.force
        if maxmov:
            dx = min(dx,maxmov)
            dy = min(dy,maxmov)
        return x+dx,y+dy

class ForceLayout(vg_layouts.GraphLayout):

    def __init__(self, graph):
        vg_layouts.GraphLayout.__init__(self, graph)

        self._f_rand = random.SystemRandom()
        self._f_randint = self._f_rand.randint
        self._f_mmax = 200          # Max delta for X or Y during ticks
        self._f_imax = None         # Max physx ticks from on call to layoutGraph()
        self._f_drag = 0.80         # Drag coeficient to slow nodes down
        self._f_charge = 80         # Used in coulomb calcs
        #self._f_minforce = 4        # When is the graph "stable enough"
        self._f_springrate = 0.1    # Used in hooke calcs
        self._f_minavgforce = 0.25

    def setMaxTickMove(self, mmax):
        '''
        Set the maximum distance that any one node is allowed
        to move in one tick of the engine. This can keep overly
        strong repulsion reactions in check.
        '''
        self._f_mmax = mmax

    def getLayoutSize(self):
        return 99999,99999

    def setRandomLayout(self, graph):

        rands = {}
        nodes = graph.getNodes()
        # give a small random layout to any node with no position
        randmax = len( nodes ) * 10
        for nid,nprops in nodes:

            if nprops.get('position'):
                continue

            rpos = ( self._f_randint( 1, randmax ), self._f_randint( 1, randmax ) )
            while rands.get(rpos):
                rpos = ( self._f_randint( 1, randmax ), self._f_randint( 1, randmax ) )

            rands[rpos] = True
            nprops['position'] = rpos


    def layoutGraph(self):

        # Setup a random layout to start with
        self.setRandomLayout( self.graph )

        # Layout each contiguous cluster seperately
        cgraphs = self.graph.getClusterGraphs()
        for graph in cgraphs:

            nodes = graph.getNodes()
            edges = graph.getEdges()

            try:

                for i in itertools.count():

                    totforce = self._tickPhysicsEngine(graph)

                    # If the system's overall energy is less than minforce, done!
                    #if totforce < self._f_minforce:
                    if totforce / len(nodes) < self._f_minavgforce:
                        break

                    if self._f_imax and i >= self._f_imax:
                        break

            except Exception:
                logger.error(traceback.format_exc())

        # Now, in order from largest to smallest, shift them back toward 0,0
        cgraphs.sort( cmp=lambda x,y: cmp(y.getNodeCount(), x.getNodeCount()) )

        offset = 0
        for graph in cgraphs:
            x,y,xmax,ymax = self._getLayoutGeom(graph)
            self._shiftGraphLayout( graph, 0 - x, offset - y )
            offset += ( ymax - y )
            #self._shiftGraphLayout( graph, offset - x, 0 - y )
            #offset += ( xmax - x )

        # Now lets calculate some straight line edges
        self._setEdgePoints()

    def _shiftGraphLayout(self, graph, dx, dy):
        nodes = graph.getNodes()
        # ugly speed hack... "ephemeral list comprehension"
        [ graph.setNodeProp(n, 'position', (n[1]['position'][0] + dx, n[1]['position'][1] + dy) ) for n in nodes ]

    def _getLayoutGeom(self, graph):
        # return x,y,xmax,ymax for this graph
        xmin = 0xffffffff
        ymin = 0xffffffff
        xmax = 0
        ymax = 0
        nodes = graph.getNodes()
        for nid,nprops in nodes:

            w,h = nprops.get('size')
            x,y = nprops.get('position')

            xmin = min( xmin, x - ( w / 2 ) )
            ymin = min( ymin, y - ( h / 2 ) )
            xmax = max( xmax, x + ( w / 2 ) )
            ymax = max( ymax, y + ( h / 2 ) )

        return xmin,ymin,xmax,ymax

    def incLayoutGraph(self):
        '''
        Layout the graph "incrementally" ( and return bool showing
        weather or not we need another increment ).

        Example:
            while l.incLayoutGraph():

        '''
        self.setRandomLayout( self.graph )
        cgraphs = self.graph.getClusterGraphs()

        needmore = False

        for graph in cgraphs:

            nodes = graph.getNodes()
            edges = graph.getEdges()

            try:

                totforce = self._tickPhysicsEngine(graph)
                if totforce / len(nodes) > self._f_minavgforce:
                    needmore = True

            except Exception as e:
                logger.error(traceback.format_exc())

        # Now, in order from largest to smallest, shift them back toward 0,0
        cgraphs.sort( cmp=lambda x,y: cmp(y.getNodeCount(), x.getNodeCount()) )

        offset = 0
        for graph in cgraphs:
            x,y,xmax,ymax = self._getLayoutGeom( graph )
            self._shiftGraphLayout( graph, 0 - x, offset - y )
            offset += ( ymax - y )
            #self._shiftGraphLayout( graph, offset - x, 0 - y )
            #offset += ( xmax - x )

        # Now lets calculate some straight line edges
        self._setEdgePoints()
        return needmore

    def _setEdgePoints(self):
        for eid,n1,n2,einfo in self.graph.getEdges():
            n1pos = self.graph.getNodeProps(n1)['position']
            n2pos = self.graph.getNodeProps(n2)['position']
            einfo['edge_points'] = [ n1pos, n2pos ]

    def _tickPhysicsEngine(self, graph):

        totforce = 0

        nodes = graph.getNodes()

        for nid, nprops in nodes:

            vect = vector()

            # For each other node, calculate the coulomb repulsion
            # as a directional vector and sum them
            vadd = vect.addVect
            repul = self._coulombRepulsion

            # Speed hack (ephemeral list comprehension)
            [ vadd( *repul(nprops, n2props ) ) for (n2id,n2props) in nodes if n2id != nid ]

            # Now lets calculate some edge "spring" forces and add
            # those in too...
            for eid,n1,n2,einfo in graph.getRefsFromByNid( nid ):
                n2props = graph.getNodeProps( n2 )
                force,angle = self._hookeAttraction( nprops, n2props )
                vect.addVect(force, angle)

            for eid,n1,n2,einfo in graph.getRefsToByNid( nid ):
                n2props = graph.getNodeProps( n1 )
                force,angle = self._hookeAttraction( nprops, n2props )
                vect.addVect(force, angle)

            # All repulsive forces are added to this node's vect
            # time to add some drag....
            drag = nprops.get('drag')
            if drag is None:
                drag = self._f_drag
            vect.addDrag( drag )
            # FIXME allow the graph to specify drag per node

            newpos = vect.getMovement( *nprops['position'], maxmov=self._f_mmax)
            graph.setNodeProp( (nid,nprops), 'position', newpos)

            totforce += vect.force

        return totforce

    def _coulombRepulsion(self, n1props, n2props ):
        x1,y1 = n1props.get('position')
        x2,y2 = n2props.get('position')

        # Calculate the deltas
        dx = x2-x1
        dy = y2-y1

        distance = math.hypot(dx, dy)
        force = ke * ( ( self._f_charge * self._f_charge ) / distance**2 )
        # Invert the angle to make the force "repulsion"
        angle = math.atan2(dy, dx) - math.pi
        return force,angle

    def _hookeAttraction(self, n1props, n2props):
        x1,y1 = n1props.get('position')
        x2,y2 = n2props.get('position')
        dx = x2-x1
        dy = y2-y1
        distance = math.hypot(dx, dy)
        angle = math.atan2(dy, dx)
        force = distance * self._f_springrate
        return force,angle




import collections
import envi.symstore.symcache as es_symcache

# Symbol Type Constants ( for serialization )
SYMSTOR_SYM_SYMBOL = 0
SYMSTOR_SYM_FUNCTION = 1
SYMSTOR_SYM_SECTION = 2
SYMSTOR_SYM_MODULE = 3


class Symbol:

    symtype = SYMSTOR_SYM_SYMBOL

    def __init__(self, name, value, size=0, fname=None):
        self.name = name
        self.value = value
        self.size = size
        self.fname = fname

    def __ge__(self, other):
        return int(self) >= int(other)
    def __le__(self, other):
        return int(self) <= int(other)
    def __gt__(self, other):
        return int(self) > int(other)
    def __lt__(self, other):
        return int(self) < int(other)
    def __eq__(self, other):
        if not isinstance(other, Symbol):
            return False
        return int(self) == int(other)
    def __add__(self, other):
        return int(self) + int(other)
    def __sub__(self, other):
        return int(self) - int(other)
    def __mul__(self, other):
        return int(self) * int(other)
    def __div__(self, other):
        return int(self) / int(other)
    def __floordiv__(self, other):
        return int(self) // int(other)
    def __mod__(self, other):
        return int(self) % int(other)
    def __divmod__(self, other):
        return divmod(int(self), int(other))
    def __pow__(self, other, modulo=None):
        return pow(int(self), int(other), modulo)
    def __lshift__(self, other):
        return int(self) << int(other)
    def __rshift__(self, other):
        return int(self) >> int(other)
    def __and__(self, other):
        return int(self) & int(other)
    def __xor__(self, other):
        return int(self) ^ int(other)
    def __or__(self, other):
        return int(self) | int(other)
    # Operator swapped variants
    def __radd__(self, other):
        return int(other) + int(self)
    def __rsub__(self, other):
        return int(other) - int(self)
    def __rmul__(self, other):
        return int(other) * int(self)
    def __rdiv__(self, other):
        return int(other) / int(self)
    def __rfloordiv__(self, other):
        return int(other) // int(self)
    def __rmod__(self, other):
        return int(other) % int(self)
    def __rdivmod__(self, other):
        return divmod(int(other), int(self))
    def __rpow__(self, other, modulo=None):
        return pow(int(other), int(self), modulo)
    def __rlshift__(self, other):
        return int(other) << int(self)
    def __rrshift__(self, other):
        return int(other) >> int(self)
    def __rand__(self, other):
        return int(other) & int(self)
    def __rxor__(self, other):
        return int(other) ^ int(self)
    def __ror__(self, other):
        return int(other) | int(self)

    # Inplace variants
    def __iadd__(self, other):
        self.value += int(other)
        return self
    def __isub__(self, other):
        self.value -= int(other)
        return self
    def __imul__(self, other):
        self.value *= int(other)
        return self
    def __idiv__(self, other):
        self.value = int(self.value / int(other))
        return self
    def __ifloordiv__(self, other):
        self.value //= int(other)
        return self
    def __imod__(self, other):
        self.vsSetValue(self % other)
        self.value %= int(other)
        return self
    def __ipow__(self, other, modulo=None):
        self.value = pow(self.value, other, modulo)
        return self
    def __ilshift__(self, other):
        self.value <<= other
        return self
    def __irshift__(self, other):
        self.value >>= other
        return self
    def __iand__(self, other):
        self.value &= other
        return self
    def __ixor__(self, other):
        self.value ^= other
        return self
    def __ior__(self, other):
        self.value |= other
        return self

    def __hash__(self):
        return hash(int(self))

    def __int__(self):
        return int(self.value)

    def __len__(self):
        return self.size

    def __str__(self):
        if self.fname is not None:
            return "%s.%s" % (self.fname, self.name)
        return self.name

    def __repr__(self):
        return str(self)


class FunctionSymbol(Symbol):
    """
    Used to represent functions.
    """
    symtype = SYMSTOR_SYM_FUNCTION

    def __repr__(self):
        return "%s.%s()" % (self.fname, self.name)


class SectionSymbol(Symbol):
    """
    Used for file sections/segments.
    """
    symtype = SYMSTOR_SYM_SECTION

    def __repr__(self):
        return "%s[%s]" % (self.fname, self.name)


class SymbolResolver:
    """
    NOTE: Nothing should reach directly into a SymbolResolver!
    """

    def __init__(self, width=4, casesens=True, baseaddr=0):
        self.width = width
        self.widthmask = (2**(width*8))-1
        self.casesens = casesens
        self.baseaddr = baseaddr    # Set if this is an RVA sym resolver

        # Lets use 4096 byte buckes for now
        self.bucketsize = 4096
        self.bucketmask = self.widthmask ^ (self.bucketsize-1)

        self.buckets = collections.defaultdict(list)

        # holds tuples by name/addr, instantiated on demand and subsequently
        # stored in symobjsbyaddr and symobjsbyname
        self.symnames = {}
        self.symaddrs = {}

        # caches that hold instantiated Symbol objects
        self.symobjsbyaddr = {}
        self.symobjsbyname = {}

    def delSymbol(self, sym):
        """
        Delete a symbol from the resolver's namespace
        """
        symval = int(sym)
        self.symaddrs.pop(symval, None)

        # bbase = symval & self.bucketmask
        # self.objbuckets[bbase].remove(sym)

        subres = None
        if sym.fname is not None:
            subres = self.symnames.get(sym.fname)

        # Potentially del it from the sub resolver's namespace
        if subres and isinstance(subres, es_symcache.SymbolCache):
            subres.delSymbol(sym)

        # Otherwise del it from our namespace
        else:
            symname = sym.name
            if not self.casesens:
                symname = symname.lower()

            if symname in self.symnames:
                self.symnames.pop(symname, None)

            if sym.fname in self.symobjsbyname:
                self.symobjsbyname.pop(sym.fname, None)

            if symval in self.symobjsbyaddr:
                self.symobjsbyaddr.pop(symval, None)

    def addSymbol(self, sym):
        """
        Add a symbol to the resolver.
        """
        # Fake these out for the API ( optimized implementations should *not* call this )
        symtup = (sym.value, sym.size, sym.name, sym.symtype, sym.fname)
        symtups = [symtup]

        self._nomSymTupAddrs(symtups)

        subres = self.symobjsbyname.get(sym.fname)
        if subres:
            subres._nomSymTupAddrs(symtups)
            subres._nomSymTupNames(symtups)
        else:
            self._nomSymTupNames(symtups)

        self._nomSymTupAddrs(symtups)

        return self._addSymObject(sym)

    def getSymByName(self, name):
        '''
        Retrieve a Symbol object by name.
        '''
        if not self.casesens:
            name = name.lower()

        # Do we have a cached object?
        sym = self.symobjsbyname.get(name)
        if sym is not None:
            return sym

        # Do we have a symbol tuple?
        symtup = self.symnames.get(name)
        if symtup is not None:
            return self._symFromTup(symtup)

    def delSymByName(self, name):
        if not self.casesens:
            name = name.lower()

        sym = self.symnames.get(name, None)
        if sym is not None:
            self.delSymbol(self._symFromTup(sym))

    def _symFromTup(self, symtup):
        # Create a symbol object and cache it...
        symaddr, symsize, symname, symtype, symfname = symtup
        symclass = symclasses[symtype]
        if symtype == SYMSTOR_SYM_MODULE:
            sym = FileSymbol(symname, symaddr, symsize, width=self.width)
        else:
            sym = symclass(symname, symaddr, size=symsize, fname=symfname)

        self._addSymObject(sym)
        return sym

    def _addSymObject(self, sym):
        # Add a symbol object to our datastructures.
        self.symobjsbyaddr[sym.value] = sym

        symmax = sym.value + sym.size

        bbase = sym.value & self.bucketmask

        if sym.fname:
            subres = self.symobjsbyname.get(sym.fname)
            if subres is not None and isinstance(subres, SymbolResolver):
                subres._addSymObject(sym)
                return

        symname = sym.name
        if not self.casesens:
            symname = symname.lower()

        self.symobjsbyname[symname] = sym

    def getSymByAddr(self, va, exact=True):
        """
        Return a symbol object for the given virtual address.
        """
        va = va & self.widthmask

        sym = self.symobjsbyaddr.get(va)
        if sym is not None:
            return sym

        symtup = self.symaddrs.get(va)
        if symtup:
            return self._symFromTup(symtup)

        # In the "not exact" case, go by the tuples...
        # ...and try 2 buckets... ( more than 8k away is bunk )
        if not exact:
            bucketva = va & self.bucketmask
            b1 = [ b for b in self.buckets[bucketva] if b[0] <= va ]
            if not b1:
                b1 = self.buckets[bucketva - self.bucketsize]

            if b1:
                b1.sort()
                symtup = b1[-1]
                sym = self.symobjsbyaddr.get(symtup[0])
                if sym is not None:
                    return sym

                return self._symFromTup(symtup)

    def getSymList(self):
        """
        Return a list of the symbols which are contained in this resolver.
        """
        out = [self.getSymByName(name) for name in self.symobjsbyname]
        out.extend([self.getSymByName(name) for name in self.symnames])
        return out

    def getSymHint(self, va, hidx):
        """
        May be used by symbol resolvers who know what type they are
        resolving to store and retrieve "hints" with indexes.

        Used specifically by opcode render methods to resolve
        any memory dereference info for a given operand.

        NOTE: These are mostly symbolic references to FRAME LOCAL
              names....
        """
        return None

    def _nomSymTupAddrs(self, symtups):

        # Ugly list comprehensions for speed...
        [self.symaddrs.__setitem__(n[0], n) for n in symtups]

        for symtup in symtups:
            # do the size range...
            self.buckets[symtup[0] & self.bucketmask].append(symtup)
            if symtup[1]:
                [self.buckets[b].append(symtup) for b in range(symtup[0], symtup[0] + symtup[1], self.bucketsize)]

    def _nomSymTupNames(self, symtups):
        if not self.casesens:
            [self.symnames.__setitem__( n[2].lower(), n ) for n in symtups]
        else:
            [self.symnames.__setitem__( n[2], n ) for n in symtups]

    def impSymCache(self, symcache, symfname=None, baseaddr=0):
        '''
        Import a list of symbol tuples (see getCacheSyms()) at the
        given base address ( and for the given sub-file )
        '''
        # Recieve a "cache" list and make it into our kind of tuples.
        symtups = [(symaddr + baseaddr, symsize, symname, symtype, symfname) for (symaddr, symsize, symname, symtype) in symcache]

        # Either way, index the addresses
        self._nomSymTupAddrs(symtups)

        if symfname:
            # If we have a sub-resolver, no need to add the names to
            # our name space...
            subres = self.symobjsbyname.get(symfname)
            if isinstance(subres, SymbolResolver):
                subres._nomSymTupAddrs(symtups)
                subres._nomSymTupNames(symtups)
                return

        self._nomSymTupNames(symtups)


class FileSymbol(Symbol, SymbolResolver):
    """
    A file symbol is both a symbol resolver of it's own, and
    a symbol.

    File symbols are used to do heirarchal symbol lookups and don't
    actually add anything but the name to their lookup (it is assumed
    that the parent Resolver of the FileSymbol takes care of addr lookups.
    """
    symtype = SYMSTOR_SYM_MODULE

    def __init__(self, fname, base, size, width=4):
        if fname is None:
            raise Exception('fname must not be None for a FileSymbol')

        SymbolResolver.__init__(self, width=width, baseaddr=base)
        Symbol.__init__(self, fname, base, size=size, fname=None)

    def __getattr__(self, name):
        """
        File symbols may be dereferenced like python objects to resolve
        symbols within them.
        """
        ret = self.getSymByName(name)
        if ret is None:
            raise AttributeError("%s has no symbol %s" % (self.name, name))
        return ret

    def __getitem__(self, name):
        """
        Allow dictionary style access for mangled incompatible names...
        """
        ret = self.getSymByName(name)
        if ret is None:
            raise KeyError("%s has no symbol %s" % (self.name, name))
        return ret

    # we need __getstate__ and __setstate__ because of serialization.  if
    # these are not overridden, __getattr__ is called, which subsequently calls
    # getSymByName, which tries to access self.casesens, which causes a
    # __getattr__ call, which leads to recursion.
    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, sdict):
        self.__dict__.update(sdict)

    # we don't *have* to override the other object methods, but otherwise
    # we will get incur the cost of extra symbol lookups for things like
    # __eq__, __ne__, etc.  we chose not to do it for lt, le, gt, ge, del and
    # others that we don't expect to see called often.
    def __repr__(self):
        return Symbol.__repr__(self)

    def __str__(self):
        return Symbol.__str__(self)

    def __eq__(self, other):
        return Symbol.__eq__(self, other)

    def __ne__(self, other):
        return not Symbol.__eq__(self, other)

    def __hash__(self):
        return Symbol.__hash__(self)

    def __nonzero__(self):
        return True

symclasses = (Symbol, FunctionSymbol, SectionSymbol, FileSymbol)



import unittest

import vivisect.symboliks.common as vsc
import vivisect.symboliks.substitution as vss


class TestSubstitution(unittest.TestCase):
    '''
    symobj = (Var('x',4) * Const(3, 4) ) + Var('y',4)
    print(str(symobj))

    x = srange('x', 3)
    y = sset('y', [1,9,0xffffffff])

    for i in (x+y):
        print('solved: %s %d' % (repr(i),symobj.solve(vals=i),))

    for i in (x*y):
        print('solved: %s %d' % (repr(i),symobj.solve(vals=i),))
    '''
    def test_basic_substitution(self):
        symobj = vsc.Var('x', 4) + vsc.Const(3,4) + vsc.Var('y', 4)
        x = vss.srange('x', 4)
        y = vss.sset('y', [1, 2, 4, 7, 99])

        #addAnswers = []
        # TODO: The nature of AddSubState feels completely wrong
        # also, x+y and y+x produce *very* different results that don't really line up
        #for idx, vals in enumerate((x + y)):
        #for idx, vals in enumerate((y + x)):
            #self.assertEqual(symobj.solve(vals=vals), addAnswers[idx])
            #print(vals, str(symobj), symobj.solve(vals=vals))

        mulAnswers = [4, 5, 6, 7,
                      5, 6, 7, 8,
                      7, 8, 9, 10,
                      10, 11, 12, 13,
                      102, 103, 104, 105]
        for vals in (x * y):
            self.assertEqual(symobj.solve(vals=vals), mulAnswers.pop(0))
        self.assertEqual(len(mulAnswers), 0)



'''
Some of the basic/universal memory renderers.
'''

import struct

import envi.common as e_common
import envi.memory as e_memory
import envi.memcanvas as e_canvas

class ByteRend(e_canvas.MemoryRenderer):

    __fmt_char__ = 'B'
    __pad__ = False

    def __init__(self, bigend=False):

        self.fmtbase = '<'
        if bigend:
            self.fmtbase = '>'

        self.width = struct.calcsize('%s%s' % (self.fmtbase, self.__fmt_char__))
        self.bformat = '%%.%dx' % (self.width * 2)
        self.pformat = '0x%s' % self.bformat

    def render(self, mcanv, va, numbytes=16):
        bigend = mcanv.getEndian()
        self.fmtbase = ('<', '>')[bigend]

        bytez = mcanv.mem.readMemory(va, numbytes)
        if self.__pad__:
            bytez = bytez.ljust(numbytes, b'\x00')

        mcanv.addVaText(self.pformat % va, va)
        mcanv.addText('  ')

        cnt = len(bytez) // self.width
        packfmt = self.fmtbase + (self.__fmt_char__ * cnt)
        for val in struct.unpack(packfmt, bytez):
            bstr = self.bformat % val
            if mcanv.mem.isValidPointer(val):
                mcanv.addVaText(bstr, val)
            else:
                mcanv.addNameText(bstr)

            mcanv.addText(' ')

        # add padding space for anything less than 16.
        # this only 'works' if your display font is monospaced.
        diff = numbytes - len(bytez)
        mcanv.addText(' ' * 3 * diff) # 2 hex chars + space
        mcanv.addText('  ')
        self.rendChars(mcanv, bytez)
        mcanv.addText('\n')

        return len(bytez)

class ShortRend(ByteRend):

    __fmt_char__ = 'H'
    __pad__ = True

class LongRend(ByteRend):

    __fmt_char__ = 'I'
    __pad__ = True

class QuadRend(ByteRend):
    __fmt_char__ = 'Q'
    __pad__ = True

def isAscii(bytez: bytes):
    bytez = bytez.split(b'\x00')[0]
    if len(bytez) < 4:
        return False, None
    for i in range(len(bytez)):
        o = bytez[i]
        if o < 0x20 or o > 0x7e:
            return False, None
    return True, bytez

def isBasicUnicode(bytez: bytes):
    bytez = bytez.split(b'\x00\x00')[0]
    if len(bytez) < 8:
        return False, None
    nonull = bytez.replace(b'\x00', b'')
    if (len(bytez) // 2) - 1 != len(nonull):
        return False, None
    return isAscii(nonull)

def getAsciiFormatted(bytez):
    is_ascii, bytez = isAscii(bytez)
    if bytez is not None:
        bytez = "'%s'" % bytez.decode('utf8')
    return bytez

def getBasicUnicodeFormatted(bytez):
    is_uni, bytez = isBasicUnicode(bytez)
    if bytez is not None:
        bytez = "u'%s'" % bytez.decode('utf8')
    return bytez

def getSymByAddrFormatted(trace, va):
    sym = trace.getSymByAddr(va, exact=False)
    if sym is not None:
        return '%s + %d' % (repr(sym), va-int(sym))
    return sym

def getFilenameFromFdFormatted(trace, va):
    for fd, ttype, bestname in trace.getFds():
        if fd == va:
            return 'HANDLE/FD?: %s' % bestname
    return None

class AutoBytesRenderer(e_canvas.MemoryRenderer):

    def __init__(self, maxrend=32, readsize=64):
        self.maxrend = maxrend
        self.readsize = readsize

    def render(self, mcanv, va):
        trace = mcanv.mem
        max_readable = trace.getMaxReadSize(va)
        if max_readable <= 0:
            return ''
        rsize = min(self.readsize, max_readable)
        bytez = trace.readMemory(va, rsize)

        prettyperm = ''
        mmap = trace.getMemoryMap(va)
        if mmap is not None:
            addr, size, perm, fname = mmap
            prettyperm = e_memory.reprPerms(perm)

        ascii_text = getAsciiFormatted(bytez)
        uni_text = getBasicUnicodeFormatted(bytez)
        sym = getSymByAddrFormatted(trace, va)
        fname = getFilenameFromFdFormatted(trace, va)

        desc = ''
        pdesc = (ascii_text, uni_text, sym, fname)
        items = [ppdesc for ppdesc in pdesc if ppdesc is not None]
        if len(items) == 1:
            desc = items[0]
        elif len(items) == 0:
            # if none match, just return the bytes.
            desc = e_common.hexify(bytez)
        elif len(items) > 1:
            # we only really expect one or none of these to match.
            desc = 'Error, multiple matches for this address!'
        else:
            raise Exception('should never get here')

        best_desc = '%s %s' % (prettyperm, desc)
        chopped_best_desc = best_desc[:self.maxrend]
        if len(best_desc) > len(chopped_best_desc):
            mcanv.addText(' %s...' % chopped_best_desc)
        else:
            mcanv.addText(' %s' % chopped_best_desc)

        return len(chopped_best_desc)



"""
Darwin Platform Module
"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import os
import ctypes
import signal
import logging
import ctypes.util as c_util

import envi.const as e_const

import vtrace
import vtrace.archs.i386 as v_i386
import vtrace.archs.amd64 as v_amd64
import vtrace.platforms.base as v_base
import vtrace.platforms.posix as v_posix


logger = logging.getLogger(__name__)
addrof = ctypes.pointer

# The OSX ptrace defines...
PT_TRACE_ME     = 0    # child declares it's being traced
PT_READ_I       = 1    # read word in child's I space
PT_READ_D       = 2    # read word in child's D space
PT_READ_U       = 3    # read word in child's user structure
PT_WRITE_I      = 4    # write word in child's I space
PT_WRITE_D      = 5    # write word in child's D space
PT_WRITE_U      = 6    # write word in child's user structure
PT_CONTINUE     = 7    # continue the child
PT_KILL         = 8    # kill the child process
PT_STEP         = 9    # single step the child
PT_ATTACH       = 10   # trace some running process
PT_DETACH       = 11   # stop tracing a process
PT_SIGEXC       = 12   # signals as exceptions for current_proc
PT_THUPDATE     = 13   # signal for thread#
PT_ATTACHEXC    = 14   # attach to running process with signal exception
PT_FORCEQUOTA   = 30   # Enforce quota for root
PT_DENY_ATTACH  = 31

# Top-level identifiers
CTL_UNSPEC  = 0        # unused
CTL_KERN    = 1        # "high kernel": proc, limits
CTL_VM      = 2        # virtual memory
CTL_VFS     = 3        # file system, mount type is next
CTL_NET     = 4        # network, see socket.h
CTL_DEBUG   = 5        # debugging parameters
CTL_HW      = 6        # generic cpu/io
CTL_MACHDEP = 7        # machine dependent
CTL_USER    = 8        # user-level
CTL_MAXID   = 9        # number of valid top-level ids

KERN_OSTYPE          = 1    # string: system version
KERN_OSRELEASE       = 2    # string: system release
KERN_OSREV           = 3    # int: system revision
KERN_VERSION         = 4    # string: compile time info
KERN_MAXVNODES       = 5    # int: max vnodes
KERN_MAXPROC         = 6    # int: max processes
KERN_MAXFILES        = 7    # int: max open files
KERN_ARGMAX          = 8    # int: max arguments to exec
KERN_SECURELVL       = 9    # int: system security level
KERN_HOSTNAME        = 10    # string: hostname
KERN_HOSTID          = 11    # int: host identifier
KERN_CLOCKRATE       = 12    # struct: struct clockrate
KERN_VNODE           = 13    # struct: vnode structures
KERN_PROC            = 14    # struct: process entries
KERN_FILE            = 15    # struct: file entries
KERN_PROF            = 16    # node: kernel profiling info
KERN_POSIX1          = 17    # int: POSIX.1 version
KERN_NGROUPS         = 18    # int: # of supplemental group ids
KERN_JOB_CONTROL     = 19    # int: is job control available
KERN_SAVED_IDS       = 20    # int: saved set-user/group-ID
KERN_BOOTTIME        = 21    # struct: time kernel was booted
KERN_NISDOMAINNAME   = 22    # string: YP domain name
KERN_DOMAINNAME      = KERN_NISDOMAINNAME
KERN_MAXPARTITIONS   = 23    # int: number of partitions/disk
KERN_KDEBUG          = 24    # int: kernel trace points
KERN_UPDATEINTERVAL  = 25    # int: update process sleep time
KERN_OSRELDATE       = 26    # int: OS release date
KERN_NTP_PLL         = 27    # node: NTP PLL control
KERN_BOOTFILE        = 28    # string: name of booted kernel
KERN_MAXFILESPERPROC = 29    # int: max open files per proc
KERN_MAXPROCPERUID   = 30    # int: max processes per uid
KERN_DUMPDEV         = 31    # dev_t: device to dump on
KERN_IPC             = 32    # node: anything related to IPC
KERN_DUMMY           = 33    # unused
KERN_PS_STRINGS      = 34    # int: address of PS_STRINGS
KERN_USRSTACK32      = 35    # int: address of USRSTACK
KERN_LOGSIGEXIT      = 36    # int: do we log sigexit procs?
KERN_SYMFILE         = 37    # string: kernel symbol filename
KERN_PROCARGS        = 38
#/* 39 was KERN_PCSAMPLES... now deprecated
KERN_NETBOOT         = 40    # int: are we netbooted? 1=yes,0=no
KERN_PANICINFO       = 41    # node: panic UI information
KERN_SYSV            = 42    # node: System V IPC information
KERN_AFFINITY        = 43    # xxx
KERN_TRANSLATE       = 44    # xxx
KERN_CLASSIC         = KERN_TRANSLATE    # XXX backwards compat
KERN_EXEC            = 45    # xxx
KERN_CLASSICHANDLER  = KERN_EXEC # XXX backwards compatibility
KERN_AIOMAX          = 46    # int: max aio requests
KERN_AIOPROCMAX      = 47    # int: max aio requests per process
KERN_AIOTHREADS      = 48    # int: max aio worker threads
KERN_PROCARGS2       = 49
KERN_COREFILE        = 50    # string: corefile format string
KERN_COREDUMP        = 51    # int: whether to coredump at all
KERN_SUGID_COREDUMP  = 52    # int: whether to dump SUGID cores
KERN_PROCDELAYTERM   = 53    # int: set/reset current proc for delayed termination during shutdown
KERN_SHREG_PRIVATIZABLE = 54    # int: can shared regions be privatized ?
KERN_PROC_LOW_PRI_IO = 55    # int: set/reset current proc for low priority I/O
KERN_LOW_PRI_WINDOW  = 56    # int: set/reset throttle window - milliseconds
KERN_LOW_PRI_DELAY   = 57    # int: set/reset throttle delay - milliseconds
KERN_POSIX           = 58    # node: posix tunables
KERN_USRSTACK64      = 59    # LP64 user stack query
KERN_NX_PROTECTION   = 60    # int: whether no-execute protection is enabled
KERN_TFP             = 61    # Task for pid settings
KERN_PROCNAME        = 62    # setup process program  name(2*MAXCOMLEN)
KERN_THALTSTACK      = 63    # for compat with older x86 and does nothing
KERN_SPECULATIVE_READS  = 64    # int: whether speculative reads are disabled
KERN_OSVERSION       = 65    # for build number i.e. 9A127
KERN_SAFEBOOT        = 66    # are we booted safe?
KERN_LCTX            = 67    # node: login context
KERN_RAGEVNODE       = 68
KERN_TTY             = 69    # node: tty settings
KERN_CHECKOPENEVT    = 70      # spi: check the VOPENEVT flag on vnodes at open time
KERN_MAXID           = 71    # number of valid kern ids
# # KERN_RAGEVNODE types
KERN_RAGE_PROC       = 1
KERN_RAGE_THREAD     = 2
KERN_UNRAGE_PROC     = 3
KERN_UNRAGE_THREAD   = 4

# # KERN_OPENEVT types
KERN_OPENEVT_PROC    = 1
KERN_UNOPENEVT_PROC  = 2

# # KERN_TFP types
KERN_TFP_POLICY      = 1

# # KERN_TFP_POLICY values . All policies allow task port for self
KERN_TFP_POLICY_DENY    = 0     # Deny Mode: None allowed except privileged
KERN_TFP_POLICY_DEFAULT = 2    # Default  Mode: related ones allowed and upcall authentication

# # KERN_KDEBUG types
KERN_KDEFLAGS        = 1
KERN_KDDFLAGS        = 2
KERN_KDENABLE        = 3
KERN_KDSETBUF        = 4
KERN_KDGETBUF        = 5
KERN_KDSETUP         = 6
KERN_KDREMOVE        = 7
KERN_KDSETREG        = 8
KERN_KDGETREG        = 9
KERN_KDREADTR        = 10
KERN_KDPIDTR         = 11
KERN_KDTHRMAP        = 12
# # Don't use 13 as it is overloaded with KERN_VNODE
KERN_KDPIDEX         = 14
KERN_KDSETRTCDEC     = 15
KERN_KDGETENTROPY    = 16

# # KERN_PANICINFO types
KERN_PANICINFO_MAXSIZE  = 1    # quad: panic UI image size limit
KERN_PANICINFO_IMAGE    = 2    # panic UI in 8-bit kraw format

# * KERN_PROC subtypes
KERN_PROC_ALL        = 0    # everything
KERN_PROC_PID        = 1    # by process id
KERN_PROC_PGRP       = 2    # by process group id
KERN_PROC_SESSION    = 3    # by session of pid
KERN_PROC_TTY        = 4    # by controlling tty
KERN_PROC_UID        = 5    # by effective uid
KERN_PROC_RUID       = 6    # by real uid
KERN_PROC_LCID       = 7    # by login context id

# Stupid backwards perms defs...
VM_PROT_READ	= 1
VM_PROT_WRITE	= 2
VM_PROT_EXECUTE	= 4

# Thread status types...
x86_THREAD_STATE32    = 1
x86_FLOAT_STATE32     = 2
x86_EXCEPTION_STATE32 = 3
x86_THREAD_STATE64    = 4
x86_FLOAT_STATE64     = 5
x86_EXCEPTION_STATE64 = 6
x86_THREAD_STATE      = 7
x86_FLOAT_STATE       = 8
x86_EXCEPTION_STATE   = 9
x86_DEBUG_STATE32     = 10
x86_DEBUG_STATE64     = 11
x86_DEBUG_STATE       = 12
THREAD_STATE_NONE     = 13

class X86_STATE_HDR(ctypes.Structure):
    _fields_ = [
        ('flavor', ctypes.c_uint32),
        ('count',  ctypes.c_uint32),
    ]

class STRUCT_X86_THREAD_STATE32(ctypes.Structure):
    _fields_ = [
        #('tsh', X86_STATE_HDR),
        ('eax', ctypes.c_uint32),
        ('ebx', ctypes.c_uint32),
        ('ecx', ctypes.c_uint32),
        ('edx', ctypes.c_uint32),
        ('edi', ctypes.c_uint32),
        ('esi', ctypes.c_uint32),
        ('ebp', ctypes.c_uint32),
        ('esp', ctypes.c_uint32),
        ('ss',  ctypes.c_uint32),
        ('eflags', ctypes.c_uint32),
        ('eip', ctypes.c_uint32),
        ('cs',  ctypes.c_uint32),
        ('ds',  ctypes.c_uint32),
        ('es',  ctypes.c_uint32),
        ('fs',  ctypes.c_uint32),
        ('gs',  ctypes.c_uint32),
    ]

class STRUCT_X86_EXCEPTION_STATE32(ctypes.Structure):
    _fields_ = [
        ('trapno',     ctypes.c_uint32),
        ('err',        ctypes.c_uint32),
        ('faultvaddr', ctypes.c_uint32),
    ]

class STRUCT_X86_DEBUG_STATE32(ctypes.Structure):
    _fields_ = [ ('debug%d', ctypes.c_uint32) for i in range(8) ]

class STRUCT_X86_THREAD_STATE64(ctypes.Structure):
    _fields_ = [
        #('tsh', X86_STATE_HDR),
        ('rax', ctypes.c_uint64),
        ('rbx', ctypes.c_uint64),
        ('rcx', ctypes.c_uint64),
        ('rdx', ctypes.c_uint64),
        ('rdi', ctypes.c_uint64),
        ('rsi', ctypes.c_uint64),
        ('rbp', ctypes.c_uint64),
        ('rsp', ctypes.c_uint64),
        ('r8',  ctypes.c_uint64),
        ('r9',  ctypes.c_uint64),
        ('r10', ctypes.c_uint64),
        ('r11', ctypes.c_uint64),
        ('r12', ctypes.c_uint64),
        ('r13', ctypes.c_uint64),
        ('r14', ctypes.c_uint64),
        ('r15', ctypes.c_uint64),
        ('rip', ctypes.c_uint64),
        ('rflags', ctypes.c_uint64),
        ('cs',  ctypes.c_uint64),
        ('fs',  ctypes.c_uint64),
        ('gs',  ctypes.c_uint64),
    ]


class STRUCT_X86_EXCEPTION_STATE64(ctypes.Structure):
    _fields_ = [
        ('trapno',     ctypes.c_uint32),
        ('err',        ctypes.c_uint32),
        ('faultvaddr', ctypes.c_uint64),
    ]

class STRUCT_X86_DEBUG_STATE64(ctypes.Structure):
    _fields_ = [ ('debug%d', ctypes.c_uint64) for i in range(8) ]

###########################################################################
#
# mach port enumerations
#
MACH_PORT_NULL              = 0

#MACH_PORT_RIGHT_* definitions are used as arguments
MACH_PORT_RIGHT_SEND        = 0
MACH_PORT_RIGHT_RECEIVE     = 1
MACH_PORT_RIGHT_SEND_ONCE   = 2
MACH_PORT_RIGHT_PORT_SET    = 3
MACH_PORT_RIGHT_DEAD_NAME   = 4
MACH_PORT_RIGHT_LABELH      = 5
MACH_PORT_RIGHT_NUMBER      = 6

def MACH_PORT_TYPE(right):
    return 1 << (right + 16)

MACH_PORT_TYPE_SEND         = MACH_PORT_TYPE(MACH_PORT_RIGHT_SEND)
MACH_PORT_TYPE_RECEIVE      = MACH_PORT_TYPE(MACH_PORT_RIGHT_RECEIVE)
MACH_PORT_TYPE_SEND_ONCE    = MACH_PORT_TYPE(MACH_PORT_RIGHT_SEND_ONCE)
MACH_PORT_TYPE_PORT_SET     = MACH_PORT_TYPE(MACH_PORT_RIGHT_PORT_SET)
MACH_PORT_TYPE_DEAD_NAME    = MACH_PORT_TYPE(MACH_PORT_RIGHT_DEAD_NAME)
MACH_PORT_TYPE_LABELH       = MACH_PORT_TYPE(MACH_PORT_RIGHT_LABELH)

###########################################################################
#
# mach message types and structures
#
MACH_MSG_TIMEOUT_NONE = 0
MACH_MSG_OPTION_NONE  = 0

MACH_SEND_MSG       = 0x00000001
MACH_RCV_MSG        = 0x00000002
MACH_RCV_LARGE      = 0x00000004

MACH_SEND_TIMEOUT   = 0x00000010
MACH_SEND_INTERRUPT = 0x00000040  # libmach implements
MACH_SEND_CANCEL    = 0x00000080
MACH_SEND_ALWAYS    = 0x00010000  # internal use only
MACH_SEND_TRAILER   = 0x00020000  

MACH_RCV_TIMEOUT    = 0x00000100
MACH_RCV_NOTIFY     = 0x00000200
MACH_RCV_INTERRUPT  = 0x00000400  # libmach implements
MACH_RCV_OVERWRITE  = 0x00001000

# Return codes from mach_msg...
MACH_RCV_TIMED_OUT  = 0x10004003

MACH_MSG_TYPE_MOVE_RECEIVE   = 16    # Must hold receive rights
MACH_MSG_TYPE_MOVE_SEND      = 17    # Must hold send rights
MACH_MSG_TYPE_MOVE_SEND_ONCE = 18    # Must hold sendonce rights
MACH_MSG_TYPE_COPY_SEND      = 19    # Must hold send rights
MACH_MSG_TYPE_MAKE_SEND      = 20    # Must hold receive rights
MACH_MSG_TYPE_MAKE_SEND_ONCE = 21    # Must hold receive rights
MACH_MSG_TYPE_COPY_RECEIVE   = 22    # Must hold receive rights

size_t            = ctypes.c_ulong

mach_port_t       = ctypes.c_uint32
mach_port_name_t  = ctypes.c_uint32
mach_port_right_t = ctypes.c_uint32
mach_msg_size_t   = ctypes.c_uint32
mach_msg_bits_t   = ctypes.c_uint32
mach_msg_id_t     = ctypes.c_uint32

ipc_space_t       = ctypes.c_uint32

kern_return_t     = ctypes.c_uint32

class mach_msg_header_t(ctypes.Structure):
    _fields_ = [
      ('msgh_bits',        mach_msg_bits_t),
      ('msgh_size',        mach_msg_size_t),
      ('msgh_remote_port', mach_port_t),
      ('msgh_local_port',  mach_port_t),
      ('msgh_reserved',    mach_msg_size_t),
      ('msgh_id',          mach_msg_id_t),
    ]

class mach_msg_body_t(ctypes.Structure):
    _fields_ = [
        ('msgh_descriptor_count', ctypes.c_uint32),
    ]

class mach_msg_port_descriptor_t(ctypes.Structure):
    _fields_ = [
        ('name',        mach_port_t),
        ('pad1',        mach_msg_size_t),
        ('pad2',        ctypes.c_uint32),
    ]

class NDR_record_t(ctypes.Structure):
    _fields_ = [
        ('mig_vers',     ctypes.c_uint8),
        ('if_vers',      ctypes.c_uint8),
        ('reserved',     ctypes.c_uint8),
        ('mig_encoding', ctypes.c_uint8),
        ('int_rep',      ctypes.c_uint8),
        ('char_rep',     ctypes.c_uint8),
        ('float_rep',    ctypes.c_uint8),
        ('reserved2',    ctypes.c_uint8),
    ]

exception_type_t        = ctypes.c_uint32
mach_msg_type_number_t  = ctypes.c_uint32
exception_data_t        = ctypes.POINTER(ctypes.c_uint32)

# the message type we receive from the kernel for exceptions
class exc_msg(ctypes.Structure):
    _fields_ = [
        ('Head',    mach_msg_header_t),
        #('data',    ctypes.c_uint8 * 1024),

        ('body',    mach_msg_body_t),
        ('thread',  mach_msg_port_descriptor_t),
        ('task',    mach_msg_port_descriptor_t),
        ('NDR',     NDR_record_t),
        ('exception', exception_type_t),
        ('codeCnt',   mach_msg_type_number_t),
        ('codes',     ctypes.c_uint32 * 128),


        ##('codes',     exception_data_t),
        ##('pad',       ctypes.c_uint8 * 512)

    ]

# The response message we send back
class exc_rep_msg(ctypes.Structure):
    _fields_ = [
        ('Head',    mach_msg_header_t),
        ('data',    ctypes.c_uint8 * 1024),
        #('NDR',     NDR_record_t),
        #('RetCode', ctypes.c_uint32)
    ]

##########################################################################
# mach generic exception codes
#
EXC_BAD_ACCESS            = 1
EXC_BAD_INSTRUCTION       = 2
EXC_ARITHMETIC            = 3
EXC_EMULATION             = 4
EXC_SOFTWARE              = 5
EXC_BREAKPOINT            = 6
EXC_SYSCALL               = 7
EXC_MACH_SYSCALL          = 8
EXC_RPC_ALERT             = 9
EXC_CRASH                 = 10

# EXC_SOFTWARE will have code[0] == EXC_SOFT_SIGNAL for posix sigs
EXC_SOFT_BREAK            = 0x00001 # LLDB: (exc_type == EXC_BREAKPOINT || ((exc_type == EXC_SOFTWARE) && exc_data[0] == 1))
EXC_SOFT_SIGNAL           = 0x10003 # Unix signal exceptions

EXC_MASK_MACHINE            = 0
EXC_MASK_BAD_ACCESS         = 1 << EXC_BAD_ACCESS
EXC_MASK_BAD_INSTRUCTION    = 1 << EXC_BAD_INSTRUCTION
EXC_MASK_ARITHMETIC         = 1 << EXC_ARITHMETIC
EXC_MASK_EMULATION          = 1 << EXC_EMULATION
EXC_MASK_SOFTWARE           = 1 << EXC_SOFTWARE
EXC_MASK_BREAKPOINT         = 1 << EXC_BREAKPOINT
EXC_MASK_SYSCALL            = 1 << EXC_SYSCALL
EXC_MASK_MACH_SYSCALL       = 1 << EXC_MACH_SYSCALL
EXC_MASK_RPC_ALERT          = 1 << EXC_RPC_ALERT
EXC_MASK_CRASH              = 1 << EXC_CRASH

EXC_MASK_ALL = (EXC_MASK_BAD_ACCESS |
                EXC_MASK_BAD_INSTRUCTION |
                EXC_MASK_ARITHMETIC |
                EXC_MASK_EMULATION |
                EXC_MASK_SOFTWARE |
                EXC_MASK_BREAKPOINT |
                EXC_MASK_SYSCALL |
                EXC_MASK_MACH_SYSCALL |
                EXC_MASK_RPC_ALERT |
                EXC_MASK_CRASH |
                EXC_MASK_MACHINE)

EXCEPTION_DEFAULT        = 1 # Send a catch_exception_raise message including the identity.
EXCEPTION_STATE          = 2 # Send a catch_exception_raise_state message including the thread state.
EXCEPTION_STATE_IDENTITY = 3 # Send a catch_exception_raise_state_identity message including the thread identity and state.
MACH_EXCEPTION_CODES     = 0x80000000 # Send 64-bit code and subcode in the exception header

boolean_t = ctypes.c_uint32
pid_t     = ctypes.c_uint32
#u_int     = ctypes.c_uint32
#pvoid     = ctypes.c_void_p
#fixpt_t   = ctypes.c_uint32
#u_quad_t  = ctypes.c_ulonglong
#sigset_t  = ctypes.c_uint32
thread_t  = ctypes.c_uint32


####################################################################
#
# mach VM related stuff....
#
vm_prot_t    = ctypes.c_uint32
vm_inherit_t = ctypes.c_uint32
vm_behavior_t = ctypes.c_uint32
memory_object_offset_t = ctypes.c_ulonglong

VM_REGION_BASIC_INFO_64    = 9

class vm_region_basic_info_64(ctypes.Structure):
    _fields_ = [
        ('protection',      vm_prot_t),
        ('max_protection',  vm_prot_t),
        ('inheritance',     vm_inherit_t),
        ('shared',          boolean_t),
        ('reserved',        boolean_t),
        ('offset',          memory_object_offset_t),
        ('behavior',        vm_behavior_t),
        ('user_wired_count',ctypes.c_ushort),
    ]

VM_REGION_BASIC_INFO_COUNT_64 = ctypes.sizeof(vm_region_basic_info_64) / 4

#mach_helper = ctypes.CDLL('./darwin_mach.dylib')

#
# These are used by the machhelper library code
#
class DarwinDebugCtx(ctypes.Structure):
    _fields_ = [
        ('dbgtask', mach_port_t),
        ('task',    mach_port_t),
        ('portset', mach_port_name_t),
        ('excport', mach_port_name_t),
        ('msgin',   ctypes.c_void_p),
        ('msgout',  ctypes.c_void_p),
    ]

class ProcessListEntry(ctypes.Structure):
    _fields_ = [
        ('pid', ctypes.c_uint),
        ('name', ctypes.c_char * 17),
    ]

darwindir = os.path.dirname(__file__)

####################################################################
    
class DarwinMixin(v_posix.PosixMixin, v_posix.PtraceMixin):

    def __init__(self):
        v_posix.PosixMixin.__init__(self)
        v_posix.PtraceMixin.__init__(self)

        self.libc = ctypes.CDLL(c_util.find_library('c'))
        self.myport = self.libc.mach_task_self()

        self.libc.mach_port_allocate.argtypes = [ipc_space_t, mach_port_right_t, ctypes.POINTER(mach_port_name_t)]
        self.libc.mach_port_allocate.restype = kern_return_t

        self.libc.mach_vm_read.argtypes = [ mach_port_t, size_t, size_t, ctypes.POINTER(ctypes.c_void_p), ctypes.POINTER(ctypes.c_uint32)]
        self.libc.mach_vm_read.restype = kern_return_t

        self.libc.ptrace.restype = ctypes.c_int
        self.libc.ptrace.argtypes = [ctypes.c_int, ctypes.c_uint32, ctypes.c_size_t, ctypes.c_int]

        machhelp_path = os.path.join(darwindir, 'machhelper.dylib')
        self.machhelper = ctypes.CDLL(machhelp_path)
        self.machhelper.platformPs.restype = ctypes.POINTER(ProcessListEntry)

        self.useptrace = False

        self.portset = self.newMachPort(MACH_PORT_RIGHT_PORT_SET)
        self.excport = self.newMachRWPort()
        self.addPortToSet(self.excport)

    def platformPs(self):

        ret = []
        y = self.machhelper.platformPs()
        i = 0
        while y[i].pid != 0xffffffff:
            ret.append((y[i].pid, y[i].name))
            i += 1

        # FIXME free!
        ret.reverse()
        return ret

    def platformParseBinary(self, filename, baseaddr, normname):
        pass

    def platformGetFds(self):
        logger.warning('platformGetFds() needs implementing on darwin')
        return []

    def platformExec(self, cmdline):
        pid = v_posix.PtraceMixin.platformExec(self, cmdline)
        self.task = self.taskForPid(pid)
        self.setExceptionPort()
        return pid

    def _getThreadPorts(self):
        count = ctypes.c_uint32()
        tlist = ctypes.POINTER(thread_t)()
        assert( self.libc.task_threads(self.task, addrof(tlist), addrof(count)) == 0 )
        ret = [ tlist[i] for i in range(count.value)]
        self.libc.vm_deallocate(self.task, tlist)
        return ret

    def platformSuspendThread(self, tid):
        self.libc.thread_suspend(tid)

    def platformResumeThread(self, tid):
        self.libc.thread_resume(tid)

    def platformGetThreads(self):
        ret = {}
        for tid in self._getThreadPorts():
            ret[tid] = tid
        return ret

    def platformGetMaps(self):
        maps = []
        address = ctypes.c_ulong(0)
        mapsize = ctypes.c_ulong(0)
        name    = ctypes.c_uint32(0)
        count   = ctypes.c_uint32(VM_REGION_BASIC_INFO_COUNT_64)
        info    = vm_region_basic_info_64()

        while True:

            r = self.libc.mach_vm_region(self.task, addrof(address),
                                   addrof(mapsize), VM_REGION_BASIC_INFO_64,
                                   addrof(info), addrof(count),
                                   addrof(name))

            # If we get told "invalid address", we have crossed into kernel land...
            if r == 1:
                break

            if r != 0:
                self.libc.mach_error("mach_vm_region", r)
                raise Exception('vm_region Failed for 0x%.8x: 0x%.8x' % (address.value,r))

            perms = 0
            p = info.protection
            if p & VM_PROT_READ:
                perms |= e_const.MM_READ
            if p & VM_PROT_WRITE:
                perms |= e_const.MM_WRITE
            if p & VM_PROT_EXECUTE:
                perms |= e_const.MM_EXEC
            if info.shared:
                perms |= e_const.MM_SHARED
            # If we got any perms, report the map
            if perms:
                maps.append((address.value, mapsize.value, perms, ''))

            address.value += mapsize.value

        return maps

    def handlePosixSignal(self, sig):

        if sig == signal.SIGTRAP:
            # FIXME I think we can catch these!
            # Traps on posix systems are a little complicated
            if self.stepping:
                self.stepping = False
                self.fireNotifiers(vtrace.NOTIFY_STEP)

            # FIXME and these too...
            elif self.checkBreakpoints():
                # It was either a known BP or a sendBreak()
                return

            elif self.execing:
                self.execing = False
                self.handleAttach()

            else:
                self._fireSignal(sig)

        elif sig == signal.SIGSTOP:
            # We get a regular POSIX stop signal on attach
            #self.attaching = False
            self.handleAttach()

        else:
            self._fireSignal(sig)
                                   
    def platformProcessEvent(self, event):
        """
        Handle a mach exception message
        """
        #if self.attaching:
            #self.useptrace = True
            #return self.handlePosixSignal(event)

        #self.useptrace = False

        # Some event states that need to be reset
        self.softexc = False

        threadid, excode, codes = event

        # Set the thread that signaled.
        self.setMeta('ThreadId', threadid)
        self.setMeta('StoppedThreadId', threadid)
        self.setMeta('MachException', event)

        if excode == EXC_SOFTWARE:

            self.softexc = True

            assert(len(codes) == 2)
            assert(codes[0] == EXC_SOFT_SIGNAL)

            sig = codes[1]
            self.handlePosixSignal(sig)

        elif excode == EXC_BAD_ACCESS:
            logger.warning('exc_bad_access %s', str([hex(x) for x in codes]))
            signo = signal.SIGSEGV
            #if codes[0] == KERN_INVALID_ADDRESS:
                #signo = signal.SIGBUS

            self._fireSignal(signo)

        elif excode == EXC_BAD_INSTRUCTION:
            logger.warning('exc_bad_instruction %s', str([hex(x) for x in codes]))
            self._fireSignal(signal.SIGILL)

        elif excode == EXC_CRASH:
            logger.warning('Crash: %s', str([hex(x) for x in codes]))
            self._fireExit(0xffffffff)

        elif excode == EXC_BREAKPOINT:
            logger.warning('exc_breakpoint: %s', str(codes))
            self.handlePosixSignal(signal.SIGTRAP)

        else:
            logger.warning('Unprocessed Exception Type: %d', excode)
            self.fireNotifiers(vtrace.NOTIFY_SIGNAL)

        return

    def platformAttach(self, pid):
        #print 'CLASSIC',self.machhelper.is_pid_classic(pid)
        #self.attaching = True
        self.task = self.taskForPid(pid)
        self.setExceptionPort()
        assert( self.libc.ptrace(PT_ATTACHEXC, pid, 0, 0) == 0 )

    def taskForPid(self, pid):
        task = ctypes.c_uint32()
        ret = self.libc.task_for_pid(self.myport, pid, addrof(task))
        if ret != 0:
            raise Exception('task_for_pid failed: 0x%.8x\n' % ret)
        return task.value

    def newMachPort(self, right):
        port = mach_port_name_t()
        assert( self.libc.mach_port_allocate(self.myport, right, addrof(port)) == 0)
        return port.value

    def newMachRWPort(self):
        port = self.newMachPort(MACH_PORT_RIGHT_RECEIVE)
        assert( self.libc.mach_port_insert_right(self.myport, port, port, MACH_MSG_TYPE_MAKE_SEND) == 0 )
        return port

    def addPortToSet(self, port):
        assert( self.libc.mach_port_move_member(self.myport, port, self.portset) == 0 )

    def setExceptionPort(self):
        # Set the target task's exception port to our excport
        #r = self.libc.task_set_exception_ports(self.task, EXC_MASK_SOFTWARE, self.excport,
        r = self.libc.task_set_exception_ports( self.task,
                                                EXC_MASK_ALL,
                                                self.excport,
                                                EXCEPTION_DEFAULT,
                                                THREAD_STATE_NONE)
        if r != 0:
            raise Exception('task_set_exception_ports failed: 0x%.8x' % r)

    def _getNextExc(self, timeout=None):

        exc = exc_msg()

        flags = MACH_RCV_MSG | MACH_RCV_LARGE
        if timeout is not None:
            flags |= MACH_RCV_TIMEOUT
            
        r = self.libc.mach_msg(addrof(exc),
                           flags,
                           0,                   # Send size...
                           ctypes.sizeof(exc),  # Recv msg size
                           self.excport,
                           timeout,
                           MACH_PORT_NULL)

        if r == MACH_RCV_TIMED_OUT:
            return None

        if r != 0:
            raise Exception('mach_msg (RECV) failed: 0x%.8x' % r)

        return exc

    def platformWait(self):
        # Wait for a mach message on the exception port

        #exc = None
        #while exc is None:
            #exc = self._getNextExc()

        #self.setMeta('ThreadId', exc.thread.name)
        #self.setMeta('StoppedThreadId', exc.thread.name)

        #e2 = self._getNextExc(timeout=0)
        #if e2 is not None:
        #print "ALSO GOT",e2


        # Sometimes there are still posix signals anyway...
        #while os.waitpid(-1, os.WNOHANG) != (0,0):
            #pass

        #if self.attaching:
            #pid,status = os.waitpid(self.pid, 0)
            #return os.WSTOPSIG(status)

        #pid,status = os.waitpid(self.pid, 0)
        exc = self._getNextExc()

        # Suspend the task so reading etc is safe...
        self.libc.task_suspend(self.task)

        # NOTE We must extract *all* needed info from the event here!
        codes = [exc.codes[i] for i in range(exc.codeCnt)]
        ret = (exc.thread.name, exc.exception, codes)

        self.setMeta('MachExcMsg', exc)
        #self.sendExcResp(exc)

        return ret

    def sendExcResp(self, exc, maskexc=False):

        res = self.buildExcResp(exc, maskexc=maskexc)
        x = self.libc.mach_msg( addrof(res),
                                MACH_SEND_MSG,
                                res.Head.msgh_size,
                                0, 
                                res.Head.msgh_remote_port,
                                MACH_MSG_TIMEOUT_NONE,
                                MACH_PORT_NULL)

        if x != 0:
            raise Exception('mach_msg MACH_SEND_MSG failed: 0x%.8x' % (x,))

    def buildExcResp(self, exc, maskexc=False):
        res = exc_rep_msg()
        self.machhelper.vtrace_exc_server(ctypes.pointer(exc.Head), ctypes.pointer(res.Head), maskexc)
        return res

    def platformStepi(self):
        self.stepping = True

        exc = self.getMeta('MachExcMsg')
        if exc is not None:
            self.setMeta('MachExcMsg', None)
            maskexc = ( self.getCurrentSignal() is None )
            self.sendExcResp(exc, maskexc=maskexc)

        assert( self.libc.task_resume(self.task) == 0 )
        assert( self.libc.ptrace(PT_STEP, self.pid, 1, 0) == 0 )

    def platformContinue(self):

        sig = self.getCurrentSignal()
        if sig is None:
            sig = 0

        #threadid = self.getMeta('StoppedThreadId', 0)
        #if self.softexc:
            #assert( self.macptrace(PT_THUPDATE, self.pid, threadid, sig) == 0 )

        # If we have a mach message to respond to, lets do that....
        exc = self.getMeta('MachExcMsg')
        if exc is not None:
            self.setMeta('MachExcMsg', None)
            maskexc = ( self.getCurrentSignal() is None )
            self.sendExcResp(exc, maskexc=maskexc)

        #if self.useptrace:
            #assert( self.libc.ptrace(PT_THUPDATE, self.pid, threadid, sig) == 0 )
            #assert( self.libc.ptrace(PT_CONTINUE, self.pid, 1, sig) == 0 )
            #return

        if self.softexc:
            assert( self.macptrace(PT_CONTINUE, self.pid, 1, sig) == 0 )

        assert( self.libc.task_resume(self.task) == 0 )

    def macptrace(self, request, pid, addr, data, zok=True):
        ret = self.libc.ptrace(request, pid, addr, data)
        if ret != 0 and zok:
            self.libc.perror('macptrace: ')
        return ret

    def platformDetach(self):
        assert( self.libc.task_resume(self.task) == 0 )
        assert( self.macptrace(PT_DETACH, self.pid, 0, 0) == 0 )

        #for threadport in self._getThreadPorts():
            #print 'threadport', self.libc.mach_port_deallocate(self.myport, threadport)

        logger.debug('darwin detach:')
        logger.debug(self.libc.mach_port_deallocate(self.myport, self.task))
        logger.debug(self.libc.mach_port_deallocate(self.myport, self.excport))
        logger.debug(self.libc.mach_port_deallocate(self.myport, self.portset))

    def platformReadMemory(self, address, size):
        pval = ctypes.c_void_p(0)
        sval = ctypes.c_uint32(0)
        assert( self.libc.mach_vm_read(self.task, address, size, addrof(pval), addrof(sval)) == 0 )
        buf = ctypes.string_at(pval.value, sval.value)
        assert( self.libc.vm_deallocate(self.myport, pval, sval) == 0 )
        return buf

    def platformWriteMemory(self, address, data):
        assert( self.libc.vm_write(self.task, address, data, len(data)) == 0 )

    # FIXME use vm_allocate for allocate memory
    # FIXME use vm_protect

class Darwini386Trace(
            vtrace.Trace,
            DarwinMixin,
            v_i386.i386Mixin,
            v_base.TracerBase):

    def __init__(self):
        vtrace.Trace.__init__(self)
        v_base.TracerBase.__init__(self)
        v_i386.i386Mixin.__init__(self)
        DarwinMixin.__init__(self)

    def getThreadException(self, tid):
        # Each arch trace must implement this...
        state = STRUCT_X86_EXCEPTION_STATE32()
        scount = ctypes.c_uint32(ctypes.sizeof(state) / 4)
        ret = self.libc.thread_get_state(tid, x86_EXCEPTION_STATE32, addrof(state), addrof(scount));
        if ret != 0:
            raise Exception('thread_get_state failed: 0x%.8x' % ret)
        return state.trapno, state.err, state.faultvaddr

    def platformGetRegCtx(self, tid):
        ctx = self.archGetRegCtx()
        # NOTE: the tid *is* the port...

        state = STRUCT_X86_THREAD_STATE32()
        scount = ctypes.c_uint32(ctypes.sizeof(state) / 4)
        ret = self.libc.thread_get_state(tid, x86_THREAD_STATE32, addrof(state), addrof(scount));
        if ret != 0:
            raise Exception('thread_get_state (THREAD_STATE32) failed: 0x%.8x' % ret)
        ctx._rctx_Import(state)

        state = STRUCT_X86_DEBUG_STATE32()
        scount = ctypes.c_uint32(ctypes.sizeof(state) / 4)
        ret = self.libc.thread_get_state(tid, x86_DEBUG_STATE32, addrof(state), addrof(scount));
        if ret != 0:
            raise Exception('thread_get_state (DEBUG_STATE32) failed: 0x%.8x' % ret)
        ctx._rctx_Import(state)

        return ctx

    def platformSetRegCtx(self, tid, ctx):
        state = STRUCT_X86_THREAD_STATE32()

        # Sync up a struct first...
        scount = ctypes.c_uint32(ctypes.sizeof(state) / 4)
        ret = self.libc.thread_get_state(tid, x86_THREAD_STATE32, addrof(state), addrof(scount));
        if ret != 0:
            raise Exception('thread_get_state (THREAD_STATE32) failed: 0x%.8x' % ret)

        # Export our shit into it...
        ctx._rctx_Export(state)

        scount = ctypes.sizeof(state) / 4
        r = self.libc.thread_set_state(tid, x86_THREAD_STATE32, addrof(state), scount)
        if r != 0:
            raise Exception('thread_set_state (THREAD_STATE32) failed: 0x%.8x' % r)

        state = STRUCT_X86_DEBUG_STATE32()
        ctx._rctx_Export(state)
        scount = ctypes.sizeof(state) / 4
        r = self.libc.thread_set_state(tid, x86_DEBUG_STATE32, addrof(state), scount)
        if r != 0:
            raise Exception('thread_set_state (DEBUG_STATE32) failed: 0x%.8x' % r)

class DarwinAmd64Trace(
            vtrace.Trace,
            DarwinMixin,
            v_amd64.Amd64Mixin,
            v_base.TracerBase):

    def __init__(self):
        vtrace.Trace.__init__(self)
        v_base.TracerBase.__init__(self)
        v_amd64.Amd64Mixin.__init__(self)
        DarwinMixin.__init__(self)

    def getThreadException(self, tid):
        # Each arch trace must implement this...
        state = STRUCT_X86_EXCEPTION_STATE64()
        scount = ctypes.c_uint32(ctypes.sizeof(state) / 8)
        ret = self.libc.thread_get_state(tid, x86_EXCEPTION_STATE64, addrof(state), addrof(scount));
        if ret != 0:
            raise Exception('thread_get_state failed: 0x%.8x' % ret)
        return state.trapno, state.err, state.faultvaddr

    def platformGetRegCtx(self, tid):
        ctx = self.archGetRegCtx()
        # NOTE: the tid *is* the port...

        state = STRUCT_X86_THREAD_STATE64()
        scount = ctypes.c_uint32(ctypes.sizeof(state) / 4)
        ret = self.libc.thread_get_state(tid, x86_THREAD_STATE64, addrof(state), addrof(scount));
        if ret != 0:
            self.libc.mach_error("thread_get_state x86_THREAD_STATE64 failed:", ret)
            raise Exception('thread_get_state (THREAD_STATE64) failed: 0x%.8x' % ret)
        ctx._rctx_Import(state)

        state = STRUCT_X86_DEBUG_STATE64()
        scount = ctypes.c_uint32(ctypes.sizeof(state) / 4)
        ret = self.libc.thread_get_state(tid, x86_DEBUG_STATE64, addrof(state), addrof(scount));
        if ret != 0:
            self.libc.mach_error("thread_get_state x86_DEBUG_STATE64 failed:", ret)
            raise Exception('thread_get_state (DEBUG_STATE64) failed: 0x%.8x' % ret)
        ctx._rctx_Import(state)

        return ctx

    def platformSetRegCtx(self, tid, ctx):

        state = STRUCT_X86_THREAD_STATE64()

        # Sync up a struct first...
        scount = ctypes.c_uint32(ctypes.sizeof(state) / 8)
        ret = self.libc.thread_get_state(tid, x86_THREAD_STATE64, addrof(state), addrof(scount));
        if ret != 0:
            raise Exception('thread_get_state (THREAD_STATE64) failed: 0x%.8x' % ret)

        # Export our shit into it...
        ctx._rctx_Export(state)

        scount = ctypes.sizeof(state) / 8
        r = self.libc.thread_set_state(tid, x86_THREAD_STATE64, addrof(state), scount)
        if r != 0:
            raise Exception('thread_set_state (THREAD_STATE64) failed: 0x%.8x' % r)

        state = STRUCT_X86_DEBUG_STATE64()
        ctx._rctx_Export(state)
        scount = ctypes.sizeof(state) / 8
        r = self.libc.thread_set_state(tid, x86_DEBUG_STATE64, addrof(state), scount)
        if r != 0:
            raise Exception('thread_set_state (DEBUG_STATE64) failed: 0x%.8x' % r)



"""
Finds functions containing specific instructions to determine if
their operands should be a  pointer locations.
This module should be run after codeblocks analyis pass.
"""

import envi
import vivisect.exc as v_exc
import vivisect.impemu.monitor as viv_imp_monitor

STOS = ('stosb', 'stosw', 'stosd', 'stosq')


class instrhook_watcher(viv_imp_monitor.EmulationMonitor):

    def __init__(self, vw, tryva):
        viv_imp_monitor.EmulationMonitor.__init__(self)
        self.vw = vw
        self.tryva = tryva
        self.hasret = False
        self.mndist = {}
        self.insn_count = 0
        self.lastop = None
        self.badcode = False
        self.badops = vw.arch.archGetBadOps()
        self.arch = vw.getMeta('Architecture')

    def prehook(self, emu, op, eip):
        if op in self.badops:
            emu.stopEmu()
            raise v_exc.BadOpBytes(op.va)

        if op.mnem in STOS:
            if self.arch == 'i386':
                reg = emu.getRegister(envi.archs.i386.REG_EDI)
            elif self.arch == 'amd64':
                reg = emu.getRegister(envi.archs.amd64.REG_RDI)
            if self.vw.isValidPointer(reg) and self.vw.getLocation(reg) is None:
                self.vw.makePointer(reg, follow=True)

def analyzeFunction(vw, fva):

    emulate = False
    dist = vw.getFunctionMeta(fva, 'MnemDist', default=[])

    for s in STOS:
        if s in dist:
            emulate = True
            break

    if emulate:
        emu = vw.getEmulator(va=fva)
        instrwat = instrhook_watcher(vw, fva)
        emu.setEmulationMonitor(instrwat)
        try:
            emu.runFunction(fva, maxhit=1)
        except Exception:
            pass



'''
Convert from one vivisect storage format to another.
'''

import os
import sys
import argparse

import vivisect
import vivisect.parsers as v_parsers

storemap = {
    'viv': 'vivisect.storage.basicfile',
    'mpviv': 'vivisect.storage.mpfile',
}


def setup():
    ap = argparse.ArgumentParser('Convert from one workspace format to another')

    ap.add_argument('old', help='Path to older workspace')
    ap.add_argument('--name', '-n', help='Name for the new workspace')

    return ap


def convert(old, newname):
    oldfmt = v_parsers.guessFormatFilename(old)
    vw = vivisect.VivWorkspace()
    if oldfmt not in storemap:
        print('Refusing to handle format %s, this is for workspace files only!')
        return -1

    newfmt = 'mpviv'
    if oldfmt == 'mpviv':
        newfmt = 'viv'

    if not newname:
        newname = '%s.%s' % (old, newfmt)

    oldstor = storemap[oldfmt]
    newstor = storemap[newfmt]

    vw.setMeta('StorageModule', oldstor)
    vw.loadWorkspace(old)

    vw.setMeta('StorageModule', newstor)
    vw.setMeta('StorageName', newname)
    vw.saveWorkspace()


def main(argv):
    opts = setup().parse_args(argv)

    old = os.path.abspath(opts.old)

    convert(old, opts.name)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import sys
import ctypes

import main

if __name__ == '__main__':
    main.waitForTest()

    buf = ctypes.create_string_buffer( 32 )
    sys.stdout.write('0x%.8x\r\n' % ctypes.addressof( buf ))
    sys.stdout.flush()

    main.safeReadline()

    sys.stdout.write('%s\n' % buf.value )
    sys.stdout.flush()

    main.exitTest()



'''
Extensions for vivisect may be implemented as python
modules which contain the function "vivExtension".

All ";" seperated directories listed in the VIV_EXT_PATH
environment variable will be searched for ".py" python
modules which implement "vivExtension"

The module's vivExtension function takes a vivisect workspace
(vw) and an vivisect gui reference (if present).
'''

import os
import sys
import logging
import importlib
import traceback

logger = logging.getLogger(__name__)

def loadExtensions(vw, vwgui):

    extdir = os.getenv('VIV_EXT_PATH')

    if extdir is None:
        # if user hasn't overridden the Extension Path, use the built-in default
        extdir = os.sep.join([vw.vivhome, 'plugins'])

    for dirname in extdir.split(os.pathsep):

        if not os.path.isdir(dirname):
            vw.vprint('Invalid VIV_EXT_PATH dir: %s' % dirname)
            continue

        if dirname not in sys.path:
            sys.path.append(dirname)

        for fname in os.listdir(dirname):
            modpath = os.path.join(dirname, fname)
            # dig into first level of directories and exec the __init__.py
            if os.path.isdir(modpath):
                modpath = os.path.join(modpath, '__init__.py')
                if not os.path.exists(modpath):
                    continue

            # otherwise, run all the .py files in the VIV_EXT_PATH dir
            elif not modpath.endswith('.py'):
                continue

            try:
                # Build code objects from the module files
                spec = importlib.util.spec_from_file_location(fname, modpath)
                module = importlib.util.module_from_spec(spec)
                module.vw = vw
                spec.loader.exec_module(module)

                if not hasattr(module, 'vivExtension'):
                    logger.warning("Skipping python module %r, not a valid extension (module has no vivExtension() function)", modpath)
                    continue

                logger.info("Loading extension: %r", modpath)
                module.vivExtension(vw, vwgui)
                vw.addExtension(fname, module)


            except Exception:
                vw.vprint('Extension Error: %s' % modpath)
                vw.vprint(traceback.format_exc())



"""
Unified config object for all vtoys.
"""

import os
import json
import logging

import envi.exc as e_exc

logger = logging.getLogger(__name__)

CONFIG_PATH = 0
CONFIG_ENTRY = 1


def gethomedir(*paths, **kwargs):
    makedir = kwargs.get('makedir', True)
    homepath = os.path.expanduser('~')
    path = os.path.join(homepath, *paths)

    if path is not None and not os.path.exists(path) and makedir:
        try:
            os.makedirs(path)
        except Exception as err:
            logger.warning('FIXME - invalid homedir, playing along... (%s)', err)

    return path


def loadConfig(defaults=None, docs=None, autosave=False, cfgdir=None):
    '''
    Loads config from a viv.json file.  If one does not exist, loads the 
    provided defaults.

    Returns: EnviConfig with appropriate config data
    '''
    if cfgdir:
        vivhome = os.path.abspath(cfgdir)
    else:
        vivhome = gethomedir(".viv", makedir=autosave)

    cfgpath = os.path.join(vivhome, 'viv.json')
    config = EnviConfig(filename=cfgpath, defaults=defaults, docs=docs, autosave=autosave)
    return config


class EnviConfig:
    '''
    EnviConfig basically works like a multi-layer dictionary that
    loads and stores config data.

    Set a config parameter:     cfg['foo'] = 'bar'
    Get a config parameter:     cfg['foo']
      or access parm using:     cfg.foo
    Multilevel:                 cfg.baz.bilbo.foo

    Create/get a subconfig:     cfg.getSubConfig('baz', add=True)
    Parse Text Config Options:  cfg.parseConfigOption("string.of.subconfigs=bar")

    Save the configuration:     cfg.saveConfigFile()
    Load the configuration:     cfg.loadConfigFile()
                                # both take optional filenames
    '''

    def __init__(self, filename=None, defaults=None, docs=None, autosave=False):
        self.cfginfo = {}
        self.cfgdocs = {}
        self.autosave = autosave
        self.filename = filename
        self.cfgsubsys = {}

        if defaults is not None:
            self.setConfigPrimitive(defaults)

        if filename is not None and os.path.isfile(filename):
            self.loadConfigFile(filename)

        if docs is not None:
            self.setDocsPrimitive(docs)

    def getOptionDoc(self, optname):
        '''
        Retrieve docs about the given option if present.

        Example:
            doc = config.getOptionDoc('woot')
            if doc is not None:
                print('woot: %s' % doc)
        '''
        return self.cfgdocs.get(optname)

    def getOptionByString(self, pathstring):
        '''
        Retrieve config options as a single dot-separated string.

        This is a convenience function for reading config options dynamically from a single
        string, which has benefits over config-object manipulation for alternative UIs or 
        plugins.

        Example:
            opt = config.getOptionByString('foo.bar.baz.info')

            where:
                info is a data option in side of baz
                baz is a subconfig of bar
                bar is a subconfig of foo
                foo is a top level subconfig
        '''
        optparts = pathstring.split('.')
        config = self
        for opart in optparts[:-1]:
            config = config.getSubConfig(opart, add=False)
            if config is None:
                raise e_exc.ConfigInvalidName(pathstring)

        optname = optparts[-1]
        if optname not in config.cfginfo:
            raise e_exc.ConfigInvalidOption(optname)

        return config[optname]

    def getConfigPaths(self):
        '''
        Return a list of tuples including: (type, valid path strings, existing value)

        'type' can be CONFIG_PATH or CONFIG_ENTRY to indicate whether the tuple
        represents a subconfig or an actual key/value pair
        '''
        paths = []
        todo = [([], self)]

        while todo:
            path, config = todo.pop()

            cfgkeys = config.keys()
            if cfgkeys:
                pathstr = '.'.join(path) + "."
                newpaths = [(CONFIG_ENTRY, "%s%s" % (pathstr, key), "%s" % (config[key])) for key in cfgkeys]
                paths.extend(newpaths)

            subnames = config.getSubConfigNames()
            if not len(subnames):
                paths.append((CONFIG_PATH, '.'.join(path), None))
                continue

            for subname in subnames:
                newpath = path[:]
                newpath.append(subname)
                newconfig = config.getSubConfig(subname, add=False)
                todo.append((newpath, newconfig))

        return paths

    def reprConfigPaths(self):
        '''
        Returns a string representation of the configuration paths/options
        and optionally values.  Useful for printing helper data.
        '''
        configpaths = self.getConfigPaths()
        out = ["Valid Config Entries:\n    "]
        reprs = ['%s = %s' % (ckey, cval) for ctype, ckey, cval in configpaths if ctype == CONFIG_ENTRY]
        out.append("\n    ".join(reprs))
        out.append("\n")

        out.append("\nValid Config Paths:\n    ")
        reprs = [ckey for ctype, ckey, cval in configpaths if ctype == CONFIG_PATH]
        out.append("\n    ".join(reprs))
        out.append("\n")
        return ''.join(out)

    def parseConfigOption(self, optstr):
        '''
        Parse a simple foo.bar.baz=<json> syntax string into
        the current config.
        '''
        if '=' not in optstr:
            raise e_exc.ConfigNoAssignment(optstr)

        optpath, valstr = optstr.split('=', 1)

        optparts = optpath.split('.')

        config = self
        for opart in optparts[:-1]:
            config = config.getSubConfig(opart, add=False)
            if config is None:
                raise e_exc.ConfigInvalidName(optpath)

        optname = optparts[-1]
        if optname not in config.cfginfo:
            raise e_exc.ConfigInvalidOption(optname)

        # json madness
        if valstr.startswith('0x'):
            valstr = str(int(valstr, 16))

        if not (valstr.startswith('"') and valstr.endswith('"')):
            if valstr.lower() == 'true':
                valstr = 'true'
            elif valstr.lower() == 'false':
                valstr = 'false'

            else:
                try:
                    int(valstr)
                except:
                    valstr = '"' + valstr + '"'

        config[optname] = json.loads(valstr)

    def getSubConfig(self, name, add=True):
        subcfg = self.cfgsubsys.get(name)
        if subcfg is None and add:
            subcfg = EnviConfig()
            self.cfgsubsys[name] = subcfg
            subcfg.autosave = self.autosave
            # Monkey patch the save method...
            subcfg.saveConfigFile = self.saveConfigFile
        return subcfg

    def getSubConfigNames(self):
        return list(self.cfgsubsys.keys())

    def setDocsPrimitive(self, docsdict):

        for key, val in docsdict.items():

            if isinstance(val, dict):
                subcfg = self.getSubConfig(key)
                subcfg.setDocsPrimitive(val)
                continue

            self.cfgdocs[key] = val

    def setConfigDefault(self, optname, optval, optdoc):
        if optname not in self.cfginfo:
            self.cfginfo[optname] = optval
        self.cfgdocs[optname] = optdoc

    def getConfigPrimitive(self):
        ret = dict(self.cfginfo)
        for subname, subcfg in self.cfgsubsys.items():
            ret[subname] = subcfg.getConfigPrimitive()
        return ret

    def setConfigPrimitive(self, cfgdict):
        for key, val in cfgdict.items():
            if isinstance(val, dict):
                subcfg = self.getSubConfig(key)
                subcfg.setConfigPrimitive(val)
                continue

            self.cfginfo[key] = val

    def saveConfigFile(self, filename=None):
        '''
        Save the config information to file.
        '''
        if filename is None:
            filename = self.filename
        base = os.path.dirname(filename)
        if not os.path.exists(base):
            try:
                os.makedirs(base)
            except Exception as err:
                logger.warning('FIXME - invalid homedir, playing along... (%s)', err)

        cfgdict = self.getConfigPrimitive()
        with open(filename, encoding='utf-8', mode='wt') as fd:
            json.dump(cfgdict, fd, indent=2)

    def loadConfigFile(self, filename=None):
        '''
        Load config info from a file.
        '''
        if filename is None:
            filename = self.filename
        with open(filename, encoding='utf-8', mode='rt') as fd:
            cfgdict = json.load(fd)
        self.setConfigPrimitive(cfgdict)

    def __getattr__(self, name):

        value = self.cfginfo.get(name)
        if value is not None:
            return value

        value = self.cfgsubsys.get(name)
        if value:
            return value

        raise AttributeError('%s has no %s' % (self.__class__.__name__, name))

    #####################################################
    # A few things so it smells kinda like a dictionary
    def __setitem__(self, key, val):
        curval = self.cfginfo.get(key)
        if curval is not None and (type(val) is not type(curval)):
            raise ValueError('%r incompatible with %r' % (val, curval))

        self.cfginfo[key] = val

        if self.autosave:
            self.saveConfigFile()

    def __getitem__(self, key, default=None):
        return self.cfginfo.get(key, default)

    def get(self, key, default=None):
        return self.cfginfo.get(key, default)

    def pop(self, key, default=None):
        return self.cfginfo.pop(key, default)
        if self.autosave:
            self.saveConfigFile()

    def keys(self):
        return self.cfginfo.keys()

    def items(self):
        return self.cfginfo.items()



import html

from PyQt5 import QtCore, QtGui, QtWebEngine, QtWebEngineWidgets
from PyQt5.QtCore import QObject, qInstallMessageHandler
from PyQt5.QtWebChannel import QWebChannel
from PyQt5.QtWebEngineWidgets import *
from PyQt5.QtWidgets import *

import envi.exc as e_exc
import envi.common as e_common
import envi.memcanvas as e_memcanvas

import envi.qt.html as e_q_html
import envi.qt.jquery as e_q_jquery

qt_horizontal = 1
qt_vertical = 2

from binascii import *
from vqt.main import *
from vqt.common import *


class LoggerPage(QWebEnginePage):
    def javaScriptConsoleMessage(self, level, msg, line, source):
        print('------------------------------------------------------------------')
        print(f'JSconsole message = {msg}; line: {line};')
        print('------------------------------------------------------------------')


class VQMemoryCanvas(e_memcanvas.MemoryCanvas, QWebEngineView):

    def __init__(self, mem, syms=None, parent=None, **kwargs):
        e_memcanvas.MemoryCanvas.__init__(self, mem=mem, syms=syms)
        QWebEngineView.__init__(self, parent=parent, **kwargs)

        self._canv_cache = None
        self._canv_curva = None
        self._canv_curtag = None
        self._canv_rend_middle = False
        self.fname = None

        # DEV: DO NOT DELETE THIS REFERENCE TO THESE
        # Otherwise they'll get garbage collected and things like double click
        # to navigate and logging won't work
        # (but pyqt5 won't throw an exception, because ugh).
        self._log_page = LoggerPage()
        self.setPage(self._log_page)
        # https://stackoverflow.com/questions/58906917/warnings-when-instantiating-qwebchannel-object-in-javascript
        # silence all the "property <foo>  of object <bar> has no notify signal messages
        qInstallMessageHandler(lambda *args: None)
        self.channel = QWebChannel()
        self.page().setWebChannel(self.channel)
        self.channel.registerObject('vnav', self)

        htmlpage = e_q_html.template.replace('{{{jquery}}}', e_q_jquery.jquery_2_1_0)
        page = self.page()
        page.setHtml(htmlpage)
        loop = QtCore.QEventLoop()
        page.loadFinished.connect(loop.quit)
        loop.exec()
        QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.ExcludeUserInputEvents | QtCore.QEventLoop.ExcludeSocketNotifiers)
        page.runJavaScript(e_q_jquery.jquery_2_1_0)
        self.forceSync()

        page.contentsSizeChanged.connect(self._frameContentsSizeChanged)

        # Allow our parent to handle these...
        self.setAcceptDrops(False)

    def forceSync(self):
        cthr = QtCore.QThread.currentThread()
        loop = QtCore.QThread.eventDispatcher(cthr)
        loop.processEvents(QtCore.QEventLoop.ExcludeUserInputEvents | QtCore.QEventLoop.ExcludeSocketNotifiers | QtCore.QEventLoop.WaitForMoreEvents)

    def renderMemory(self, va, size, rend=None):

        if self._canv_rend_middle:
            vmap = self.mem.getMemoryMap(va)
            if vmap is None:
                raise e_exc.InvalidAddress(va)

            origva = va
            va, szdiff = self._loc_helper(max(va - size, vmap[0]))
            size += size + szdiff

            def callScroll(data):
                self._selectVa(origva)
            ret = e_memcanvas.MemoryCanvas.renderMemory(self, va, size, rend=rend, cb=callScroll)
        else:
            ret = e_memcanvas.MemoryCanvas.renderMemory(self, va, size, rend=rend)

        return ret

    def _frameContentsSizeChanged(self, size):
        if self._canv_scrolled:
            page = self.page()
            page.runJavaScript('window.scrollTo(1, 0x0fffffff);')

    @idlethread
    def _scrollToVa(self, va, cb=None):
        # TODO: Does anybody call this?
        page = self.page()
        selector = 'viv:0x%.8x' % va
        js = f'''
        var nodes = document.getElementsByName("{selector}");
        if (nodes != null && nodes.length > 0) {{
            nodes[0].scrollIntoView()
        }}
        '''
        if cb:
            page.runJavaScript(js, cb)
        else:
            page.runJavaScript(js)

    @idlethread
    def _selectVa(self, va, cb=None):
        page = self.page()
        js = '''
        selectva("0x%.8x");
        scrolltoid("a_%.8x");
        ''' % (va, va)
        if cb:
            page.runJavaScript(js, cb)
        else:
            page.runJavaScript(js)

    def _beginRenderMemory(self, va, size, rend):
        self._canv_cache = ''

    def _endRenderMemory(self, va, size, rend, cb=None, sel=None):
        self._appendInside(self._canv_cache, cb, sel)
        self._canv_cache = None

    def _beginRenderVa(self, va, cb=None, sel=None):
        self._add_raw('<a name="viv:0x%.8x" id="a_%.8x">' % (va, va), cb)

    def _endRenderVa(self, va, cb=None, sel=None):
        self._add_raw('</a>', cb)

    def _beginUpdateVas(self, valist, cb=None):
        self._canv_cache = ''

        page = self.page()

        selector = 'a#a_%.8x' % valist[0][0]
        js = f'''
        node = document.querySelector("{selector}");
        node.outerHTML = '<update id="updatetmp"></update>' + node.outerHTML;
        '''

        for va, size in valist:
            selector = 'a#a_%.8x' % va
            js += f'''
            var node = document.querySelector("{selector}");
            if (node != null) {{
                node.remove();
            }}
            '''

        if cb:
            page.runJavaScript(js, cb)
        else:
            page.runJavaScript(js)

    def _endUpdateVas(self, cb=None):
        self._canv_cache = self._canv_cache.replace('`', r'\`')
        js = f'''
        var node = document.querySelector('update#updatetmp');
        if (node != null) {{
            node.outerHTML = `{self._canv_cache}` + node.outerHTML;
        }} else {{
            console.log("Failed to grab selector of update#updatetmp")
        }}
        '''
        if cb:
            self.page().runJavaScript(js, cb)
        else:
            self.page().runJavaScript(js)

        self._canv_cache = None

    def _beginRenderPrepend(self):
        self._canv_cache = ''
        self._canv_ppjump = self._canv_rendvas[0][0]

    def _endRenderPrepend(self, cb=None, sel=None):
        if sel is None:
            sel = '#memcanvas'
        selector = 'viv:0x%.8x' % self._canv_ppjump
        self._canv_cache = self._canv_cache.replace('`', r'\`')
        js = f'''
        var node = document.querySelector("{sel}");
        node.innerHTML = `{self._canv_cache}` + node.innerHTML

        var snode = document.getElementsByName("{selector}");
        if (snode != null && snode.length > 0) {{
            snode[0].scrollIntoView()
        }}
        '''
        self._canv_cache = None
        if cb:
            self.page().runJavaScript(js, cb)
        else:
            self.page().runJavaScript(js)

    def _beginRenderAppend(self):
        self._canv_cache = ''

    def _endRenderAppend(self, cb=None, sel=None):
        if sel is None:
            sel = '#memcanvas'
        page = self.page()
        self._canv_cache = self._canv_cache.replace('`', r'\`')
        js = f'''
        document.querySelector("{sel}").innerHTML += `{self._canv_cache}`;
        '''
        self._canv_cache = None
        if cb:
            page.runJavaScript(js, cb)
        else:
            page.runJavaScript(js)

    def getNameTag(self, name, typename='name'):
        '''
        Return a "tag" for this memory canvas.  In the case of the
        qt tags, they are a tuple of html text (<opentag>, <closetag>)
        '''
        clsname = 'envi-%s' % typename
        namehex = e_common.hexify(name.lower())
        subclsname = 'envi-%s-%s' % (typename, namehex)
        return ('<span class="%s %s" envitag="%s" envival="%s" onclick="nameclick(this)">' % (clsname,subclsname,typename,namehex), '</span>')

    def getVaTag(self, va):
        # The "class" will be the same that we get back from goto event
        return ('<span class="envi-va envi-va-0x%.8x" va="0x%.8x" ondblclick="vagoto(this)" oncontextmenu="vaclick(this)" onclick="vaclick(this)">' % (va,va), '</span>')

    @QtCore.pyqtSlot(str)
    def _jsGotoExpr(self, expr):
        # The routine used by the javascript code to trigger nav events
        if self._canv_navcallback:
            self._canv_navcallback(expr)

    @QtCore.pyqtSlot(str)
    def _jsSetCurVa(self, vastr):
        self._canv_curva = int(str(vastr), 0)
        # if we click on a VA, we no longer want a tag to show up 
        # (could do this another way so we keep "last tag" but indicate which has preference??)
        self._canv_curtag = None

    @QtCore.pyqtSlot(str, str)
    def _jsSetCurTag(self, tagname, tagval):
        print("_jsSetCurTag(%r, %r)" % (tagname, tagval))
        self._canv_curtag = (str(tagname), unhexlify(str(tagval)))

    # NOTE: doing append / scroll seperately allows render to catch up
    @idlethread
    def _appendInside(self, text, cb=None, sel=None):
        if sel is None:
            sel = '#memcanvas'
        page = self.page()
        text = text.replace('`', r'\`')
        js = f'''
        document.querySelector("{sel}").innerHTML += `{text}`;
        '''
        if cb:
            page.runJavaScript(js, cb)
        else:
            page.runJavaScript(js)

    def _add_raw(self, text, cb=None, sel=None):
        # If we are in a call to renderMemory, cache til the end.
        if self._canv_cache is not None:
            self._canv_cache += text
            return

        self._appendInside(text, cb, sel)

    def addText(self, text, tag=None, cb=None):
        text = html.escape(text).encode('unicode_escape').decode('utf-8')
        if tag is not None:
            otag, ctag = tag
            text = otag + text + ctag
        self._add_raw(text, cb)

    @idlethreadsync
    def clearCanvas(self, cb=None, sel=None):
        if sel is None:
            sel = '#memcanvas'
        page = self.page()
        js = f'''
        var node = document.querySelector("{sel}");
        if (node != null) {{
            node.innerHTML = "";
        }}
        '''
        if cb:
            page.runJavaScript(js, cb)
        else:
            page.runJavaScript(js)

    def contextMenuEvent(self, event):
        va = self._canv_curva
        tag = self._canv_curtag
        menu = QMenu()
        if self._canv_curva is not None:
            self.initMemWindowMenu(va, tag, menu)

        viewmenu = menu.addMenu('view   ')
        viewmenu.addAction("Save frame to HTML", ACT(self._menuSaveToHtml))

        menu.exec_(event.globalPos())

    def initMemWindowMenu(self, va, tag, menu):
        initMemSendtoMenu('0x%.8x' % va, menu)

    def dumpHtml(self, data):
        if self.fname:
            with open(self.fname, 'w') as f:
                f.write(data)
            self.fname = None

    def _menuSaveToHtml(self):
        fname = getSaveFileName(self, 'Save As HTML...')
        if fname is not None:
            fname = str(fname)
            if len(fname):
                self.fname = fname
                self.page().toHtml(self.dumpHtml)


def getNavTargetNames():
    '''
    Returns a list of Memory View names.
    Populated by vqt in a seperate thread, thus is time-sensitive.  If the
    list is accessed too quickly, some valid names may not yet be inserted.
    '''
    ret = []
    vqtevent('envi:nav:getnames', ret)
    return ret


def initMemSendtoMenu(expr, menu):
    navtgts = getNavTargetNames()
    navtgts.sort()
    for name in navtgts:
        args = (name, expr, None)
        menu.addAction('sendto: %s' % name, ACT(vqtevent, 'envi:nav:expr', args))
